

import os, re, random;
import subprocess as sp

from MatrixTools import *
from RQuiver import *


class DoubleBraidClusterA:
	"""A class containing and computing double braid quivers in type A.
	See [GLSBS22]=[arXiv:2210.04778] and [GLSB23]=[arXiv:2301.07268].
	Code by Pavel Galashin, posted 01/16/23.

	DISCLAIMER:
	This code was written over the course of 4 years. It is messy and redundant and possibly incorrect. Use at your own risk.
	"""

	def __init__(self,wred,ured="w0", u=None,n=None,lazy=False,debug=False,
				compute_matrices=True,use_xy_vars=True):
		"""
			INPUT:

			 - ``wred`` -- double braid word in the alphabet 1,..,n-1,-1,...,-n+1.

			 - ``n`` -- number of strands (default: determined by ``wred``).

			 - ``ured`` -- either a reduced word (on positive letters), or "w0" or "id". Must be less than or equal the Demazure product of ``wred``.

			 - ``lazy``: if ``False``, will automatically run ``self.compute_data``, computing the quivers and everything else.

			 - ``compute_matrices``: if ``True``, will compute the X_c, Y_c, Z_c matrices.

			 - ``use_xy_vars``: whether to name the variables by ``x_i,y_i`` or ``T_i``.

			 - ``reverse_u``: if True, then ``ured`` is reversed, so that positive letters in ``ured`` form an honest subword of ``wred``
			 """
		wred_abs = [abs(j) for j in wred]
		if n==None:
			if len(wred_abs)==0: n=1
			else: n = max(wred_abs) + 1
		self.n=n
		self.mt = MatrixTools(n=n)
		mt=self.mt
		self.w0 = mt.w0()
		self.id_perm = mt.id_perm()
		self.use_xy_vars=use_xy_vars
		if u==None:
			if ured == "w0":
				u = self.w0
			elif ured == "id":
				u = self.id_perm
			else:
				u = mt.from_red(ured)
				if u.length() != len(ured): raise Exception("u is not reduced! ured=%s, len=%s"%(ured,u.length()))
		self.u=u
		self.ured=ured
		self.wlen=self.m=len(wred)
		self.wred=wred
		self.wred_abs=wred_abs
		try:
			NON_TRIVIAL_MUTATIONS
		except NameError:
			NON_TRIVIAL_MUTATIONS = []
		if not lazy: self.compute_data(debug=debug,compute_matrices=compute_matrices);


	def compute_data(self,debug=False, compute_matrices=False):
		"""Compute the quiver and all the accompanying data."""
		wred,u,wred_abs,n,mt,w0,id_perm,wlen=self.wred,self.u,self.wred_abs,self.n,self.mt,self.w0,self.id_perm,self.wlen
		if debug: print("computing data: u=%s, wred=%s, n=%d"%(u,wred,n));
		nsymb="%"+str(3*n-3)+"s";
		m = wlen
		u_prev = u
		hollow = []
		solid = []
		for i in reversed(range(m)):
			wi, wia = wred[i], abs(wred[i])
			if wi < 0:
				u_next = mt.s[wia] * u_prev
			else:
				u_next = u_prev * mt.s[wia]
			if u_next.length() < u_prev.length():
				hollow.append(i)
				u_prev = u_next
			else:
				solid.append(i)
		hollow, solid = list(reversed(hollow)), list(reversed(solid))
		if debug: print("hollow+1=%s\nsolid+1=%s" % ([j + 1 for j in hollow], [j + 1 for j in solid]))
		if u_prev != id_perm:
			print("u_(0)=%s" % u_prev)
			raise Exception("no subword for u!")


		terms = []
		uprod_p = id_perm #product of positive (=top) u's
		uprod_m = id_perm #product of negative (=bottom) u's
		wprod_p = id_perm #product of  positive w's
		wprod_m = id_perm #product of negative w's

		for i in range(m):
			wi, wia = wred[i], abs(wred[i])
			if wi < 0:
				wprod_m *= mt.s[wia]
			else:
				wprod_p *= mt.s[wia]
			if i in hollow:
				terms.append(1)
				if wi < 0:
					uprod_m *= mt.s[wia]
				else:
					uprod_p *= mt.s[wia]
			else:
				ti = var("t" + str(i + 1))
				if wi < 0:
					utm = mt.dot(uprod_m)
				else:
					utp = mt.dot(uprod_p)
		if compute_matrices:
			#Propagate from the right, right-multiplying by z and z-bar matrices.
			#initial condition: X_m=U_+, Y_m=~(dot of w_0u) * U_+
			# MX=[Matrix.identity(n)]
			# MY=[~mt.dot(self.w0*self.u)]
			MX=[mt.dot(self.w0*self.u)]
			MY=[Matrix.identity(n)]
			MX_gen=[MX[0]]
			MY_gen=[MY[0]]
			use_xy_vars=self.use_xy_vars
			cnt_x=m
			cnt_y=m
			MX_red=copy(MX)
			MY_red=copy(MY)
			for i in reversed(range(m)):
				wi,wia=wred[i],abs(wred[i])
				if wi>0:
					MY=[MY[0]]+MY
					MY_gen=[MY_gen[0]]+MY_gen
					# vname=str("x"+str(cnt_x)) if use_xy_vars else "t"+str(i+1)
					vname="x"+str(i+1) if use_xy_vars else "t"+str(i+1)
					cnt_x-=1
					if i in hollow: MX0=MX[0]*mt.dot_s(wi)
					else: MX0=MX[0]*mt.z(wi,var(vname))
					MX0_gen=MX_gen[0]*mt.z(wi,var(vname))
					MX_gen= [mt.auto_simplify_matrix(MX0_gen)] + MX_gen
					MX=[MX0]+MX
					MX_red=[MX0]+MX_red
				else:
					MX=[MX[0]]+MX
					MX_gen=[MX_gen[0]]+MX_gen
					# vname=str("y"+str(cnt_y)) if use_xy_vars else "t"+str(i+1)
					vname=str("y"+str(i+1)) if use_xy_vars else "t"+str(i+1)
					cnt_y-=1
					wi2=n-wia
					if i in hollow: MY0=MY[0]*~mt.dot_s(wi2)
					else: MY0=MY[0]*mt.zb(wi2,var(vname))
					MY0_gen=MY_gen[0]*mt.zb(wi2,var(vname))
					MY_gen= [mt.auto_simplify_matrix(MY0_gen)] + MY_gen
					MY=[MY0]+MY
					MY_red=[MY0]+MY_red
			MZ=[~MY[i]*MX[i] for i in range(m+1)]
			MZ_gen=[mt.auto_simplify_matrix(~MY_gen[i] * MX_gen[i]) for i in range(m + 1)]
			self.MX,self.MX_red,self.MY,self.MY_red,self.MZ, self.MX_gen, self.MY_gen, self.MZ_gen=\
				MX,MX_red,MY,MY_red,MZ, MX_gen, MY_gen, MZ_gen

		vk = {0: id_perm} #top partial products
		wk = {0: id_perm} #bottom partial products
		uk = {0: id_perm} #u-partial products. top=+ acts on RIGHT, bottom=- on LEFT
		ukp = {0: id_perm} #top u-partial products. top acts on RIGHT
		ukm = {0: id_perm} #bot u-partial products. bottom on LEFT
		for i in range(m):
			wi, wia = wred[i], abs(wred[i])
			if i not in hollow: uk[i + 1] = uk[i]; ukp[i + 1] = ukp[i]; ukm[i + 1] = ukm[i];
			if wi > 0:
				vk[i + 1] = vk[i] * mt.s[wia]
				wk[i + 1] = wk[i]
				if i in hollow: uk[i + 1] = uk[i] * mt.s[wia]; ukp[i + 1] = ukp[i] * mt.s[wia]; ukm[i+1]=ukm[i]
			else:
				vk[i + 1] = vk[i]
				wk[i + 1] = mt.s[wia] * wk[i]
				if i in hollow: uk[i + 1] = mt.s[wia] * uk[i]; ukm[i + 1] = mt.s[wia] * ukm[i]; ukp[i+1]=ukp[i]

		current_left=[-1]*(n+1);
		left_neigh={};
		right_neigh={}
		for i in range(wlen):
			wi=abs(wred[i]);
			left_neigh[i]=current_left[wi];
			right_neigh[current_left[wi]]=i;
			current_left[wi]=i;
		for wi in range(1,n):
			if current_left[wi]==-1: continue
			right_neigh[current_left[wi]]=oo

		self.hollow, self.solid, self.uprod_p, self.uprod_m, self.wprod_p, self.wprod_m, self.vk,self.wk,self.uk,self.ukp,self.ukm, \
			self.left_neigh, self.right_neigh = \
			hollow, solid, uprod_p, uprod_m, wprod_p, wprod_m, vk,wk,uk,ukp,ukm, \
			 left_neigh, right_neigh

		self.compute_grid_labels(debug=debug)
		self.compute_quiver(debug=debug)
		if debug: print("computed double data! wred=%s, \nhollow+1=%s; solid+1=%s"%(wred,[j+1 for j in hollow],[j+1 for j in solid]))

	def get_aps(self, disturb_index=-1):
		"""  Find almost positive subexpression for ``u``.
		 ``disturb_index`` tells us where the length goes down. """
		u_prev, wred, m, mt = self.u, self.wred, self.m, self.mt

		hollow = []
		solid = []
		uk = {m: u_prev}
		for i in reversed(range(m)):
			wi, wia = wred[i], abs(wred[i])
			if wi < 0:
				u_next = mt.s[wia] * u_prev
			else:
				u_next = u_prev * mt.s[wia]
			if (u_next.length() < u_prev.length()) or (disturb_index == i and (u_next.length() > u_prev.length())):
				hollow.append(i)
				u_prev = u_next
			else:
				solid.append(i)
			uk[i] = u_prev
		hollow, solid = list(reversed(hollow)), list(reversed(solid))
		return hollow, solid, uk

	def compute_grid_labels(self, debug=False):
		""" Compute how grid minors factor into irreducibles. Computes (``i=1,2``):

		 - ``u_lab_i``: original labels of the regions by subsets

		 - ``u_lab_dis_i``: disturbed labels for each ``r`` in ``self.solid``

		 - ``F_lab_i``: factorization for each grid minor (given by two coords ``i,h``) into irreducibles  """
		n, u, wred, m, mt, uk = self.n, self.u, self.wred, self.wlen, self.mt, self.uk
		uik = {k: ~uk[k] for k in uk}
		uks_dis = {}
		uiks_dis = {}
		for r in self.solid:
			hollow_dis, solid_dis, uk_dis = self.get_aps(disturb_index=r)
			uks_dis[r] = uk_dis
			uiks_dis[r] = {k: ~uk_dis[k] for k in uk_dis}
		u_lab_1 = {}
		u_lab_2 = {}
		for i in range(m + 1):
			for h in range(1, n):
				# u_lab_1[i, h] = sorted([uk[i](j + 1) for j in range(h)])
				# u_lab_2[i, h] = sorted([uik[i](j + 1) for j in range(h)])
				u_lab_1[i, h] = sorted([uik[i](j + 1) for j in range(h)]) #sage has wrong convention for perm's
				u_lab_2[i, h] = sorted([uk[i](j + 1) for j in range(h)])
		u_lab_dis_1 = {}
		u_lab_dis_2 = {}
		for r in self.solid:
			ul1, ul2 = {}, {}
			ukd, uikd = uks_dis[r], uiks_dis[r]
			for i in range(m + 1):
				for h in range(1, n):
					# ul1[i, h] = sorted([ukd[i](j + 1) for j in range(h)])
					# ul2[i, h] = sorted([uikd[i](j + 1) for j in range(h)])
					ul1[i, h] = sorted([uikd[i](j + 1) for j in range(h)]) #sage has wrong convention for perm's
					ul2[i, h] = sorted([ukd[i](j + 1) for j in range(h)])
			u_lab_dis_1[r] = ul1
			u_lab_dis_2[r] = ul2
		F_lab_1 = {(i, h): [] for i in range(m + 1) for h in range(1, n)}
		F_lab_2 = {(i, h): [] for i in range(m + 1) for h in range(1, n)}
		for i in range(m+1):
			F_lab_1[i, 0] = F_lab_2[i, 0] = F_lab_1[i, n] = F_lab_2[i, n]=[]

		for r in self.solid:
			lab1, lab2 = u_lab_dis_1[r], u_lab_dis_2[r]
			for i in range(m + 1):
				for h in range(1, n):
					if lab1[i, h] != u_lab_1[i, h]: F_lab_1[i, h].append(r)
					if lab2[i, h] != u_lab_2[i, h]: F_lab_2[i, h].append(r)
		self.u_lab_1, self.u_lab_2, self.u_lab_dis_1, self.u_lab_dis_2, self.F_lab_1, self.F_lab_2, self.uks_dis, self.uiks_dis = \
			u_lab_1, u_lab_2, u_lab_dis_1, u_lab_dis_2, F_lab_1, F_lab_2, uks_dis, uiks_dis


	def print_grid_labels(self):
		""" Assuming ``self.compute_grid_labels()`` has been invoked.  """

		F_lab_1, F_lab_2, n, m = self.F_lab_1, self.F_lab_2, self.n, self.wlen
		for h in reversed(range(1, n)):
			labels = [[j + 1 for j in F_lab_1[i, h]] for i in range(m + 1)]
			print("row=%d; labels top=%s" % (h, "|".join(["%8s" % (lab) for lab in labels])))
		print()
		for h in reversed(range(1, n)):
			labels = [[j + 1 for j in F_lab_2[i, h]] for i in range(m + 1)]
			print("row=%d; labels bot=%s" % (h, "|".join(["%8s" % (lab) for lab in labels])))

	def grid_minor(self,c,h,return_sets=False):
		""" ``\\Delta_{c,h}`` in [GLSBS22]."""
		Z=self.MZ_gen[c]
		uc=self.uk[c]
		w0=self.w0
		if h>0:
			w0uci=~(w0*uc)
			A,B=sorted([w0uci(j+1)-1 for j in range(h)]),list(range(h))
		else:
			A,B=sorted([w0(j+1)-1 for j in range(-h)]), sorted([uc(j+1)-1 for j in range(-h)])
		mnr=self.mt.mnr(Z,A,B).factor()
		if return_sets: return (A,B,mnr)
		else: return mnr

	def check_grid_minors_tnn(self,print_negative=True):
		""" check that each \\Delta_{c,h} has no minus signs """
		n,m=self.n,self.m
		for c in range(m+1):
			for h in range(1,n):
				A1,B1,mnr1=self.grid_minor(c,h,return_sets=True)
				A2,B2,mnr2=self.grid_minor(c,-h,return_sets=True)
				if ("-" in str(mnr1) and "-" in str(1/mnr1)):
					if print_negative:
						print("negative minor! c=%d, h=%d, A=%s, B=%s, minor=%s"%(c,h,
									[j+1 for j in A1],[j+1 for j in B1],mnr1))
					return False
				if ("-" in str(mnr2) and "-" in str(1/mnr2)):
					if print_negative:
						print("negative minor! c=%d, h=%d, A=%s, B=%s, minor=%s"%(c,-h,
									[j+1 for j in A2],[j+1 for j in B2],mnr2))
					return False
		return True

	def compute_quiver(self,debug=False):
		""" Compute the quiver:

		 - ``self.frozen_list``: list of frozen vertices

	 	- ``self.quiver``: the quiver as a dictionary ``(i,j)``:[nonnegative] number of arrows ``i->j``
		 """
		F_lab_1, F_lab_2, n, m, wred = self.F_lab_1, self.F_lab_2, self.n, self.wlen, self.wred
		frozens={}
		for h in range(1,n):
			for r in F_lab_1[0,h]+F_lab_2[0,h]:
				frozens[r]=True
		frozen_list=self.frozen_list=sorted(list(frozens))
		# frozen_list=self.frozen_list=list(reversed(sorted(list(frozens))))
		region_pairs=[]
		half_arrows={}
		for i in range(m):
			wi,wia=wred[i],abs(wred[i])
			if wi>0:
				if F_lab_1[i, wia + 1] != F_lab_1[i + 1, wia + 1]:
					raise Exception(
						"top_1 labels not same! i=%d, %s != %s" % (i, F_lab_1[i, wia + 1], F_lab_1[i + 1, wia + 1]))
				if F_lab_1[i, wia - 1] != F_lab_1[i + 1, wia - 1]:
					raise Exception(
						"bot_1 labels not same! i=%d, %s != %s" % (i, F_lab_1[i, wia - 1], F_lab_1[i + 1, wia - 1]))
				chL,chR,chT,chB=F_lab_1[i, wia],F_lab_1[i+1, wia], F_lab_1[i, wia+1], F_lab_1[i, wia-1],
				region_pairs+=[(chL,chT),(chT,chR),(chL,chB),(chB,chR),(chR,chL),(chR,chL)]
			else:
				if F_lab_2[i, wia + 1] != F_lab_2[i + 1, wia + 1]:
					raise Exception(
						"top_2 labels not same! i=%d, %s != %s" % (i, F_lab_2[i, wia + 1], F_lab_2[i + 1, wia + 1]))
				if F_lab_2[i, wia - 1] != F_lab_2[i + 1, wia - 1]:
					raise Exception(
						"bot_2 labels not same! i=%d, %s != %s" % (i, F_lab_2[i, wia - 1], F_lab_2[i + 1, wia - 1]))
				chL,chR,chT,chB=F_lab_2[i, wia],F_lab_2[i+1, wia], F_lab_2[i, wia+1], F_lab_2[i, wia-1],
				region_pairs+=[(chT,chL),(chR,chT),(chB,chL),(chR,chB),(chL,chR),(chL,chR)]
		for A,B in region_pairs:
			for r1 in A:
				for r2 in B:
					if (r1,r2) not in half_arrows:
						half_arrows[r1,r2]=half_arrows[r2,r1]=0
					half_arrows[r1,r2]+=1
					half_arrows[r2,r1]-=1
		quiver={}
		for r1,r2 in half_arrows:
			narr=half_arrows[r1,r2]
			if narr<=0: continue
			if (r1 in frozen_list) and (r2 in frozen_list): continue
			if narr>2: print("wred,ured=%s,%s\nLOTS OF ARROWS! r1+1=%d, r2+1=%d, multiplicity=%d"%(
				wred, self.mt.get_red(self.u),r1+1,r2+1,narr//2)) #example with multi arrows:
			# wred, ured = [5, -2, 2, -3, 2, -4, 6, 4, -4, 5, 1, 1, 4, 4, 4, 3, 2, 1, 5, 4, 3, 3, 2, 5, 4, 6, 5, 4, 3, 2,
			# 			  1, 1, 2, 4, 5, 6, 5, 5, 5], [6, 4, 5, 6, 3, 4, 5, 2, 3, 4, 5, 6, 1, 2, 3, 4, 5, 6]  # mult arr
			# if narr>2: print("LOTS OF ARROWS! r1=%d, r2=%d, narr=%d"%(r1+1,r2+1,narr))
			if narr%2!=0: raise Exception("odd number of half-arrows! r1=%d, r2=%d, narr=%d"%(r1+1,r2+1,narr))
			quiver[(r1,r2)]=(narr//2)
		self.quiver=quiver

	def get_RQuiver(self,debug=False,index_add=0,ignore_frozens=False):
		"""Return ``self.quiver`` as an object of class ``RQuiver``."""
		solid, frozen_list, quiver= self.solid, self.frozen_list, self.quiver
		remap={};
		remap_i={};
		cnt=0;
		for i in solid:
			if i not in frozen_list:
				remap[i+index_add]=cnt;
				remap_i[cnt]=i+index_add;
				cnt+=1;
		for i in reversed(frozen_list):
			remap[i+index_add]=cnt;
			remap_i[cnt]=i+index_add;
			cnt+=1;
		if debug: print("solid=%s, frozen_list=%s, remap=%s"%(solid,frozen_list,remap));
		n_fro=m=len(frozen_list)
		n_mut=n=len(solid)-m
		B=Matrix(n+m,n);
		for i,j in quiver:
			ri=remap[i+index_add];
			rj=remap[j+index_add];
			if rj<n: B[ri,rj]+=1;
			if ri<n: B[rj,ri]-=1;
		B.set_immutable();
		return RQuiver(B,labels=remap_i,ignore_frozens=ignore_frozens)

	def get_curves(self, indices=None, debug=False, colors=None, grid_lw=0.5, grid_color="black!30",
				   axes_font_scale=0.5,point_scale=0.3,curve_lw=1.5,tikz_scale=0.5):
		"""Compute the relative cycles. (To be printed to LaTeX.)
		``indices``: which indices to print. By default, print all."""
		if colors == None:
			colors = ["red", "purple", "brown", "cyan", "blue", "orange", "green!80!black", "darkgray"]  # +list(colors);
		clen = len(colors)
		n=self.n
		m=self.wlen
		ukd=self.uks_dis
		P=""
		if indices==None: indices=range(m+1)
		names_to_ind={}
		cnt=0
		for r in self.solid:
			names_to_ind[r+1]=cnt
			cnt+=1
		for i in indices:
			u=self.uk[i]
			vs_dic = {ukd[r][i]: r + 1 for r in self.solid}
			u_list = [(u(j), j) for j in range(1, n + 1)]
			P+="\\begin{tikzpicture}[scale=%.2f]\n"%(tikz_scale)
			P += "\\draw[line width=%.2fpt,draw=%s] (1,1) grid (%.2f,%.2f);\n"%(grid_lw,grid_color,n,n)
			for j in range(1, n + 1):
				P += "\\node[scale=%.2f,blue,anchor=north] (A) at (%.2f,1) {$%d$};\n" % (axes_font_scale, j, j)
				P += "\\node[scale=%.2f,red,anchor=east] (A) at (1,%.2f) {$%d$};\n" % (axes_font_scale, j, j)
			if i>0:
				i_col="red" if self.wred[i-1]>0 else "blue"
				style="" if i-1 in self.solid else "draw=black,rectangle,"
				P+="\\node[scale=%.2f,%s,%s anchor=south east] (A) at (%.2f,%.2f) {$\\xrightarrow{%d}$};\n" % (axes_font_scale,
																							   i_col, style, 0.5, (n+1)/2, self.wred[i-1])
			cnt = 0
			if debug: print("===== i=%d ====="%i)
			for v in sorted(vs_dic, key=lambda x: vs_dic[x]):
				printed=False
				name = vs_dic[v]
				if debug: print("=== r=%d, v=%s"%(name,self.mt.get_red(v)))
				col = colors[names_to_ind[name] % clen]
				v = v
				if debug: print("u=%s; v=%s" % (u, v))
				cyc_perm = v * ~u
				for cyc in cyc_perm.cycle_tuples():
					if len(cyc) == 1: continue
					printed = True
					if debug: print("cyc=", cyc)
					jmin, jmax = min(cyc), max(cyc)
					if cyc[0] != jmin:
						raise Exception("cycle =%s should start with min=%d" % (cyc, jmin))
					below = []
					above = []
					for j in cyc:
						if j == jmin:
							mode = below
							continue
						if j == jmax:
							mode = above
							continue
						mode.append(j)
					if debug: print("below: %s, above: %s" % (below, above))
					cyc_s = sorted(list(cyc))  # sorted([(u(j),j) for j in cyc])
					xmin, ymin = jmin, u(jmin)
					xmax, ymax = jmax, u(jmax)
					verts = [(ymin, xmin), (ymax, xmax)]
					x = var("x")
					fx = ymin + (x - xmin) * (ymax - ymin) / (xmax - xmin)
					plots = {(xmin, xmax): fx}
					fxj = {j: (fx, xmin, ymin, xmax, ymax) for j in below + above}
					relax = True
					while relax:
						relax = False
						if debug: print("fxj=%s; \nplots=%s" % (fxj, plots))
						for j in fxj:
							#if debug: print("j=%d" % j)
							xj, yj = j, u(j)
							fx, x0, y0, x1, y1 = fxj[j]
							new_x, new_y = None, None
							if j in below and fx({x: xj}) >= yj:
								new_x, new_y = xj, yj - 0.5  # we flip x and y
							if j in above and fx({x: xj}) <= yj:
								new_x, new_y = xj, yj + 0.5
							if new_x == None: continue
							if debug: print("j=%d, x0=%.2f, new_x=%.2f, x1=%.2f"%(j,x0,new_x,x1))
							verts = sorted(verts + [(new_y, new_x)])
							fx0 = y0 + (x - x0) * (new_y - y0) / (new_x - x0)
							fx1 = new_y + (x - new_x) * (y1 - new_y) / (x1 - new_x)
							if (x0, x1) not in plots:
								print("""
								x0=%d, x1=%d, plots=%s
								wred,ured=%s,%s""" % (x0, x1, plots,self.wred,self.mt.get_red(self.u)))
							plots.pop((x0, x1))
							plots[x0, new_x] = fx0
							plots[new_x, x1] = fx1
							fxj_next = {}
							for k in fxj:
								if k == j: continue
								xk, yk = k, u(k)
								if x0 < xk < new_x:
									fxj_next[k] = (fx0, x0, y0, new_x, new_y)
								elif new_x < xk < x1:
									fxj_next[k] = (fx1, new_x, new_y, x1, y1)
								else:
									if debug: print("k=%d -- didn't change!"%(k))
									fxj_next[k] = fxj[k]
							fxj = fxj_next
							relax = True
							break
					# P+="\\draw[line width=%.2fpt, draw=%s, rounded corners=5] %s ;\n"%(curve_lw,col," -- ".join(["(%.2f,%.2f)"%vert for vert in verts]))
					P+="\\draw[line width=%.2fpt, draw=%s, rounded corners=5] %s ;\n"%\
					   (curve_lw,col," -- ".join(["(%.2f,%.2f)"%(vert[1],vert[0]) for vert in verts]))
				if printed:
					cnt+=1
					P += "\\node[scale=%.2f,%s,anchor=west,draw=%s,rectangle] (A) at (%.2f,%.2f) {$%d$};\n" % (axes_font_scale, col,col, n,1+cnt/1.8,name)
			for j in range(1, n + 1):
				P += "\\node[scale=%.2f,fill=black,draw=black,circle] (u%d) at (%.2f,%.2f) {};\n" % \
					 (point_scale, j, j, u(j))
			P+="\\end{tikzpicture}\n "
		return P



	@staticmethod
	def compile_latex(text, file_name="output", tex_dir="tex/", template_path="tex/double_template_A.tex",
					  error_from=500, error_to=500, return_compil_output=False, clean_aux=True, use_a1_paper=False):
		"""Output ``text`` into a latex file ``file_name``.tex using ``template_dir/template_path`` template file and compile.

		INPUT:
		 - ``error_from`` and ``error_to`` determine neighborhood size of an error message to output when error occurs.

		 - ``template_path`` is path to template file. Must contain "%PASTE%" where we will paste our new text
		output file is found in ``tex_dir/`` + ``file_name``.tex
		"""
		f=open(template_path,"r");
		template=f.read();
		subbed=template.replace("%PASTE%",text);
		if use_a1_paper:
			subbed=subbed.replace("usepackage[margin=1in]{geometry}","usepackage[a1paper,margin=1in]{geometry}")
		fname=tex_dir+file_name+".tex"
		o=open(fname,"w");
		o.write(subbed);
		o.close()
		command='latexmk -pdf -f -g -bibtex -synctex=1 -halt-on-error -jobname=%s%s %s%s.tex'%(tex_dir,file_name,tex_dir,file_name)

		pipe = sp.Popen( command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE )
		res = pipe.communicate()
		compil=res[0].decode("utf-8")

		if "error" in compil or "Error" in compil or pipe.returncode!=0:
			print("latex compilation error! return code=",pipe.returncode)
			for m in re.finditer('[Ee]rror', compil):
				print("*************************************\n")
				print(compil[max(0,m.start()-error_from):min(len(compil),m.end()+error_to)])
		# else: print("OK! file name="+file_name)
		else: html("OK! <a href='%s'>%s</a>"%(tex_dir+file_name+".pdf",file_name+".pdf"))
		if clean_aux:
			for ext in ["*.aux","*.log","*.fdb_latexmk","*.fls","*.out","*.blg","*.nav","*.snm","*.toc","*.synctex.gz"]:
				command="rm "+tex_dir+ext
				pipe = sp.Popen( command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE )
				res = pipe.communicate()
		if return_compil_output: return compil

	def get_tikzpicture(self, omit_trail=True, index_add=0, right_aligned=True, debug=False, compute_matrices=True,
						grid_label_scale=None, print_red_factorizations=True, print_blue_factorizations=True, use_set_labels=False,
						draw_quiver=True, only_mutable_quiver=False, F_shift=0, special=-1,
						print_special_arrows=True, print_curves=True, curves_indices=None,
						print_mutable_arrows=False, print_arrows=False,
						print_red_grid_minors=False, print_blue_grid_minors=False, grid_minor_scale=0.5, grid_minor_rotate=5,
						xscale=1, yscale=1):
		"""Outputs LaTeX source as a tikzpicture representing our wiring diagram.
		See the documentation for ``self.print_and_compile()`` for input descriptions."""
		wred = self.wred;
		n, hollow, solid, left_neigh, right_neigh, wred = self.n, self.hollow, self.solid, self.left_neigh, self.right_neigh, self.wred
		vk, wk, uk, w0, mt, wlen = self.vk, self.wk, self.uk, self.w0, self.mt, self.wlen
		F_lab_1,F_lab_2, u_lab_1,u_lab_2=self.F_lab_1,self.F_lab_2, self.u_lab_1,self.u_lab_2
		if omit_trail: wlen = max(solid) + 1
		m=wlen
		xscale *= min(0.61, 0.61 * 12 / wlen);
		yscale *= min(0.8, 0.8 * 6 / n)
		if grid_label_scale==None:
			grid_label_scale=min(0.5,0.5*20/wlen)
		grid_minor_scale=grid_label_scale*grid_minor_scale
		tpic = "";
		tpic += ("\\begin{tikzpicture}[xscale=%.2f,yscale=%.2f") % (xscale, yscale) + "]\n"
		tpic += ("\\def\\xsh{%.2f}\n" % (0.5 / xscale * 0.4));
		tpic += ("\\def\\ysh{%.2f}\n" % (0.8 / yscale * 0.08));
		extra_rb = 1 if right_aligned else 2;
		tpic += "\\def\\rbound{" + str(wlen * 2 + extra_rb) + "}\n"
		tpic += "\\def\\rboundOverTwo{" + str(wlen + extra_rb // 2) + "}\n"
		tpic += "\\def\\grid_label_scale{" + str(grid_label_scale) + "}\n"
		tpic += "\\wiresAB{" + str(n) + "}\n";
		i_list = list(range(wlen)) if right_aligned else reversed(list(range(wlen)))
		for i in i_list:
			wi, wia = wred[i], abs(wred[i])
			h_coord = str(i)
			neigh = -2
			if right_aligned:
				if left_neigh[i] >= 0:
					lene = str(left_neigh[i])
					neigh = left_neigh[i]
				else:
					lene = "\\lboundOverTwo";
			else:
				if right_neigh[i] != oo:
					lene = str(right_neigh[i])
					neigh = right_neigh[i]
				else:
					lene = "\\rboundOverTwo";
			# if compute_matrices:
			# 	Fs_str_list = Fs_str[i]
			# else:
			# 	Fs_str_list = " "
			# if abs(neigh - i) <= 1: Fs_str_list = " $\\scalebox{0.7}{$%s$}$ " % (Fs_str_list)
			Fs_str_list = " "
			if i not in hollow:
				if wi > 0:
					tpic += "\\dotSEPfilledA{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
				else:
					tpic += "\\dotSEPfilledB{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
			else:
				if wi > 0:
					tpic += "\\xdotSEPfilledA{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
				else:
					tpic += "\\xdotSEPfilledB{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
		lab_1,lab_2=(F_lab_1,F_lab_2) if not use_set_labels else (u_lab_1,u_lab_2)
		sep="\\!\\cdot\\!" if not use_set_labels else ","
		extra_add=1 if not use_set_labels else 0

		for i in range(m+1):
			for h in range(1,n):
				if print_red_factorizations:
					lab = sep.join([str(r+extra_add+index_add) for r in lab_1[i, h]])
					if (not use_set_labels) and (special in [r+1 for r in lab_1[i, h]]):
						tpic += "\\draw[red!40,fill=brown!40] (%.2f,%.2f) circle (0.25cm);\n" % (2*i-0.5, h+0.7)
					tpic += "\\node[red,scale=%.2f] (F1_%d_%d) at (%.2f,%.2f) {$%s $};\n" % (grid_label_scale, i, h, 2*i-0.5, h+0.7, lab)
				if print_blue_factorizations:
					lab = sep.join([str(r+extra_add+index_add) for r in lab_2[i, h]])
					if (not use_set_labels) and (special in [r+1 for r in lab_2[i, h]]):
						tpic += "\\draw[red!40,fill=brown!40] (%.2f,%.2f) circle (0.25cm);\n" % (2*i-0.5, h+0.3)
					tpic += "\\node[blue, scale=%.2f] (F1_%d_%d) at (%.2f,%.2f) {$%s $};\n" % (grid_label_scale, i, h, 2*i-0.5, h+0.3, lab)

		# print("compute_matrices=%s, print_red_grid_minors=%s"%(compute_matrices, print_red_grid_minors))
		Z=self.MZ
		set_to_minor=lambda S: "".join([str(j+1) for j in S])
		Gscale=grid_minor_scale
		for c in range(m+1):
			for h in range(1,n):
				if print_red_grid_minors:
					A1,B1,mnr1=self.grid_minor(c,h,return_sets=True)
					lab="\\Delta_{%s,%s}(Z_{%d})=%s"%(set_to_minor(A1),set_to_minor(B1),c,latex(mnr1))
					tpic += "\\node[red,rotate=%s,scale=%.2f] (G1_%d_%d) at (%.2f,%.2f) {$%s $};\n" % (grid_minor_rotate,Gscale, c, h, 2*c-0.5, h+0.7, lab)
				if print_blue_grid_minors:
					A2,B2,mnr2=self.grid_minor(c,-h,return_sets=True)
					lab="\\Delta_{%s,%s}(Z_{%d})=%s"%(set_to_minor(A2),set_to_minor(B2),c,latex(mnr2))
					tpic += "\\node[blue,rotate=%s,scale=%.2f] (G2_%d_%d) at (%.2f,%.2f) {$%s $};\n" % (grid_minor_rotate,Gscale, c, h, 2*c-0.5, h+0.3, lab)


# 			##################
		if draw_quiver:
			for r in self.solid:
				wi,wia=wred[r],abs(wred[r])
				if r in self.frozen_list:fill="black!15"
				else: fill="green!50"
				if wi>0: draw="red!70!black"; ysh="-\\shiftAB"
				else: draw="blue!70!black"; ysh="+\\shiftAB"

				tpic+="\\node[black,draw=%s,%s,fill=%s,rectangle,scale=%.2f] (F%d) at (%.2f,%.2f%s) {$%d$};\n"%\
					  (draw,draw, fill, grid_label_scale, r, 2*r+0.5-F_shift, wia+0.5,ysh, r+1+index_add)
			for x, y in self.quiver:
				if only_mutable_quiver and (x in self.frozen_list or y in self.frozen_list): continue;
				mult = self.quiver[(x, y)];
				xs,ys=str(x),str(y)
				if mult == 1:
					if (x in self.frozen_list or y in self.frozen_list):
						dred = "drawedgeEXP";
					else:
						dred = "drawedgeTHICK"
					tpic += "\\" + dred + "{" + xs + "}{" + ys + "}{" + self.get_color(x + index_add,
																					   y + index_add) + "}\n";
				else:
					tpic += "\\drawedgeMULT{" + xs + "}{" + ys + "}{" + self.get_color(x + index_add,
																					   y + index_add) + "}{" + str(
						mult) + "}\n";


		tpic += "\\end{tikzpicture}\n";
		if print_special_arrows and special!=-1:
			spec_arrows_in=[]
			spec_arrows_out=[]
			spec_arrows_mult=[]
			for x,y in self.quiver:
				if self.quiver[x,y]==1:
					if x+1 == special: spec_arrows_out.append("%d\\to%d"%(x+1,y+1))
					if y+1 == special: spec_arrows_in.append("%d\\to%d"%(x+1,y+1))
				else:
					if special in [x+1,y+1]: spec_arrows_mult.append("%d\\to%d: %d"%(x+1,y+1,self.quiver[x,y]))
			tpic+="\nspecial=$%d$: arrows in=$[%s]$; arrows out=$[%s]$"%(special,",".join(spec_arrows_in),",".join(spec_arrows_out))
			if len(spec_arrows_mult)>0: tpic+=";  MULT ARROWS=$[%s]$"%(",".join(spec_arrows_mult))
		if print_mutable_arrows or print_arrows:
			arrs=[]
			for i,j in self.quiver:
				if self.quiver[i,j]<0: continue
				if (not print_arrows) and (i in self.frozen_list or j in self.frozen_list): continue
				ii,jj=i+1+index_add,j+1+index_add
				arr="$%d\\to %d$"%(ii,jj) if self.quiver[i,j]==1 else \
					"$%d\\xrightarrow{%d}%d$"%(ii,self.quiver[i,j],jj)
				arrs.append(arr)
			title="Quiver" if print_arrows else "Mutable quiver"
			tpic+="\n\n%s: "%(title,)+", ".join(arrs)
		if print_curves:
			tpic+="\n\n"+self.get_curves(indices=curves_indices)
		return tpic + "\n\n\n"

	def get_color(self,x,y):
		"""Get color, where ``x,y`` are indices of quiver vertices (I think)."""
		return str((max(x,y)*71)%100);

	def print_grid_minors(self, print_red_grid_minors=True, print_blue_grid_minors=True, grid_minor_scale=1, xscale=3, yscale=1,
						  compact=False):
		"""Returns a tikzpicture with a wiring diagram whose regions are labeled by red/blue grid minors."""
		wred = self.wred;
		n, hollow, solid, left_neigh, right_neigh, wred = self.n, self.hollow, self.solid, self.left_neigh, self.right_neigh, self.wred
		vk, wk, uk, w0, mt, wlen = self.vk, self.wk, self.uk, self.w0, self.mt, self.wlen
		F_lab_1,F_lab_2, u_lab_1,u_lab_2=self.F_lab_1,self.F_lab_2, self.u_lab_1,self.u_lab_2
		m=wlen
		Z=self.MZ_gen
		set_to_minor=lambda S: "".join([str(j+1) for j in S])
		Gscale=grid_minor_scale
		tpic = "\\def\\gridndscale{%.2f}"%grid_minor_scale;
		tpic += ("\\begin{tikzpicture}[xscale=%.2f,yscale=%.2f") % (xscale,yscale) + "]\n"

		for c in range(m+1):
			for h in range(1,n):
				if print_red_grid_minors:
					A1,B1,mnr1=self.grid_minor(c,h,return_sets=True)
					if not compact:
						lab="\\Delta_{%s,%s}(Z_{%d})=%s"%(set_to_minor(A1),set_to_minor(B1),c,latex(mnr1))
					else:
						lab="%s"%(latex(mnr1))
					tpic += "\\node[red,scale=\\gridndscale] (G1_%d_%d) at (%.2f,%.2f) {$%s $};\n" % (c, h, c+0.5, h+0.5, lab)
		tpic+="\\end{tikzpicture}\n\n\n "
		tpic += ("\\begin{tikzpicture}[xscale=%.2f,yscale=%.2f") % (xscale,yscale) + "]\n"
		for c in range(m+1):
			for h in range(1,n):
				if print_blue_grid_minors:
					A2,B2,mnr2=self.grid_minor(c,-h,return_sets=True)
					lab="\\Delta_{%s,%s}(Z_{%d})=%s"%(set_to_minor(A2),set_to_minor(B2),c,latex(mnr2))
					tpic += "\\node[blue,scale=\\gridndscale] (G2_%d_%d) at (%.2f,%.2f) {$%s $};\n" % ( c, h, c+0.5, h+0.5, lab)
		tpic+="\\end{tikzpicture}\n "
		return tpic

	def mat_to_latex(self,M):
		"""Convert a matrix ``M`` into LaTeX."""
		Mltx=latex(M)
		return "\\smat{%s}"%"\n".join(Mltx.split("\n")[1:-1])

	def print_matrices(self, XYZ_matrix_scale=0.4):
		"""Print ``X_c,Y_c,Z_c`` matrices. Returns a tikz-cd diagram. """
		wred = self.wred;
		n, hollow, solid, left_neigh, right_neigh, wred = self.n, self.hollow, self.solid, self.left_neigh, self.right_neigh, self.wred
		vk, wk, uk, w0, mt, wlen = self.vk, self.wk, self.uk, self.w0, self.mt, self.wlen
		F_lab_1,F_lab_2, u_lab_1,u_lab_2=self.F_lab_1,self.F_lab_2, self.u_lab_1,self.u_lab_2
		m=wlen
		X,Y,Z=self.MX_gen,self.MY_gen,self.MZ_gen
		row1=""
		row2=""
		row3=""
		ltx=self.mat_to_latex
		rescale=lambda S : "\\text{\\scalebox{\\XYscale}{$%s$}}"%S
		rescale2=lambda S : "\\text{\\scalebox{\\arrscale}{$%s$}}"%S
		H={}
		for c in range(m+1):
			if c>0:
				wi,wia=wred[c-1],abs(wred[c-1])
				red_arr="\\id" if wi<0 else "s_{%d}"%wia
				red_arr=rescale2(red_arr)
				row1+="""& \\arrow[l,"\\textcolor{red}{%s}"'] %s \n"""%(red_arr,rescale("X_{%d}=%s"%(c,ltx(X[c]))))
			if c<m:
				wi,wia=wred[c],abs(wred[c])
				arr="\\id" if wi>0 else "s_{%d^\\ast}"%wia
				arr=rescale2(arr)
				row2+=""" \\arrow[r,"\\textcolor{blue}{%s}"'] %s """%(arr,rescale("Y_{%d}=%s"%(c,ltx(Y[c]))))

			if c==0:
				# row1+="X_0=%s"%ltx(X[0])
				# row1+=rescale(ltx(X[0]))
				row1+=rescale("X_0=%s\n"%ltx(X[0]))
				row2+="""\\arrow[u, Rightarrow, "w_0 \\cdot \\id"'] &"""
			elif c<m: row2+=" & \n "
			row3+=rescale("Z_{%d}=%s"%(c,ltx(Z[c])))
			if c<m: row3+="\n & \n"
			if c==m:
				# row2+="& %s "%(rescale("Y_{%d}=%s"%(c,ltx(Y[c]))))
				row1+="\\\\ "
				row2+=""" %s \\arrow[u, "w_0 \\cdot u"'] \\\\  """%(rescale("Y_{%d}=%s"%(c,ltx(Y[c]))))

		return """
		\\def\\XYscale{%.2f}
		\\def\\arrscale{0.7}
		\\begin{tikzcd}[column sep=2pt, row sep=15pt]
		%s
		%s
		%s
		\\end{tikzcd}\n\n"""%(XYZ_matrix_scale,row1,row2,row3)


	@staticmethod
	def print_and_compile(wred, ured="w0", file_name="output", right_aligned=True, name="",
						  comment="", new_page=False, return_tex=False, omit_trail=False, debug=False, grid_label_scale=None,
						  print_uw=True, index_add=0, compute_matrices=True, intro="",
						  print_red_factorizations=True, print_blue_factorizations=True,
						  draw_quiver=True, use_set_labels=False, special=-1,
						  print_curves=True, curves_indices=None,
						  print_mutable_arrows=False, print_arrows=False,
						  print_red_grid_minors=False, print_blue_grid_minors=False, grid_minor_scale=0.5,
						  use_xy_vars=True, print_grid_minors=False, print_matrices=False,
						  XYZ_matrix_scale=0.4,use_a1_paper=None):
		"""Print everything to ``file_name``.tex. Compile it if ``return_tex=False``, otherwise, return LaTeX source.

		INPUT:
		 - ``wred/ured`` -- same as for this class.
		 - ``file_name`` -- name of the output file
		 - ``name`` -- title of the section
		 - ``comment`` -- text above the wiring diagram picture
		 - ``right_aligned`` -- whether to center labels inside wiring diagram regions or make them right-aligned
		 - ``new_page`` -- whether to add \\newpage before
		 - ``omit_trail`` -- if ``True``, hollow crossings at the right end of the wiring diagram are omitted
		 - ``print_uw`` -- whether to print the reduced words for ``u`` and ``w``
		 - ``index_add`` -- increase/decrease all indices by some integer. Useful when chopping off letters from the beginning of ``wred``.
		 - ``compute_matrices`` -- whether to compute the ``X_c,Y_c,Z_c`` matrices
		 -  ``intro`` -- LaTeX text to be added at the beginning
		 - ``print_red/blue_factorizations`` -- whether to print how each red/blue grid minor factors into cluster vars
		 - ``draw_quiver`` -- whether to draw the arrows
		 - ``use_set_labels`` -- whether to label the regions by the corresponding sets of rows/columns
		 - ``special`` -- index of a cluster variable that will be highlighted in the wiring diagram
		 - ``print_curves`` -- whether to print the curves = intersections of the disks ``D_d`` with each time=c plane.
		 - ``curves_indices`` -- which curves to print (default: all)
		 - ``print_mutable_arrows`` -- if ``True``, print separately a list of all arrows between mutable vertices
		 - ``print_arrows`` -- whether to separately print a lit of all arrows (useful when the wiring diagram has too much going on)
		 - ``print_red/blue_grid_minors`` -- whether to print the actual grid minor values. Will set ``compute_matrices=True``
		 - ``grid_minor_scale`` -- rescale the grid minors so that they fit inside the regions
		 - ``use_xy_vars`` -- whether to use ``x_i,y_i`` or ``t_i`` variables
		 - ``XYZ_matrix_scale`` -- scale of ``X_c,Y_c,Z_c`` matrices.
		 - ``use_a1_paper`` -- whether to switch from letter paper to A1 paper. By default, becomes ``True`` when ``compute_matrices=True``
		 """
		tpics = "";
		if new_page: tpics += "\\newpage"
		ld = DoubleBraidClusterA(wred=wred, ured=ured, compute_matrices=compute_matrices,
								 use_xy_vars=use_xy_vars)
		if name != None: tpics += "\\section{%s}\n\n" % (name)
		wred_strs = []
		m = len(wred)
		hollow, solid = ld.hollow, ld.solid
		if print_uw:
			for i in range(m):
				wi, wia = wred[i], abs(wred[i])
				if wi > 0:
					color = "red"
				else:
					color = "blue"
				letter = "\\textcolor{%s}{%s}" % (color, wi)
				if i in ld.hollow: letter = "\\boxed{%s}" % letter
				wred_strs.append(letter)
			wred_str = "[%s]" % (",".join(wred_strs))
			tpics += "\\begin{align*} u&=\\textcolor{black}{%s}\\\\ w&=%s \\end{align*} \n\n" % (ld.ured, wred_str)
		if comment != "": tpics += "\n %s \n\n" % comment
		tpics += ld.get_tikzpicture(right_aligned=right_aligned, grid_label_scale=grid_label_scale,
							 omit_trail=omit_trail, debug=debug, compute_matrices=compute_matrices,
							 index_add=index_add,  print_red_factorizations=print_red_factorizations,
							 print_blue_factorizations=print_blue_factorizations,draw_quiver=draw_quiver,use_set_labels=use_set_labels,
							 special=special, print_curves=print_curves,curves_indices=curves_indices,
							 print_mutable_arrows=print_mutable_arrows,print_arrows=print_arrows,
							 print_red_grid_minors=print_red_grid_minors,print_blue_grid_minors=print_blue_grid_minors,
							 grid_minor_scale=grid_minor_scale)
		if print_grid_minors:
			tpics+=ld.print_grid_minors(print_red_grid_minors=print_red_grid_minors,print_blue_grid_minors=print_blue_grid_minors)
		if print_matrices:
			tpics+=ld.print_matrices(XYZ_matrix_scale=XYZ_matrix_scale)
		if use_a1_paper==None: use_a1_paper=compute_matrices
		if return_tex:
			return tpics;
		else:
			DoubleBraidClusterA.compile_latex(tpics, file_name=file_name, use_a1_paper=use_a1_paper)



	@staticmethod
	def print_multiple(datas,file_name="output",intro="",use_a1_paper=False,**opts):
		"""Print several wiring diagrams into a single file.

		INPUT:
		 - ``datas``=a list of dictionaries of quivers to print.
		each item of ``datas`` has to contain a "ured" and a "wred" parameters.
		Additionally, it may contain ``file_name``, ``name``, ``new_page``, etc.
		as in the ``self.print_and_compile()`` function. Default values for all can be passed into ``opts``.

		See the documentation for ``self.print_and_compile()`` for input descriptions."""
		global prev_quiver_local;
		prev_quiver_local = None;
		tpics=intro;
		if not isinstance(datas,list):
			if isinstance(datas,dict): datas=[datas]
			else: raise Exception("expected a list, got: %s"%datas)
		# print("intro=",intro)
		if "template_path" in opts:
			template_path=opts["template_path"]
			opts.pop("template_path")
		else:
			template_path="tex/double_template_A.tex"
		for dat in datas:
			dat2=copy(opts);
			dat2.update(dat);
			dat2["return_tex"]=True;
			if "intro" in dat2: dat2.pop("intro")
			tpics+=DoubleBraidClusterA.print_and_compile(**dat2);
		DoubleBraidClusterA.compile_latex(tpics, file_name=file_name, template_path=template_path,use_a1_paper=use_a1_paper)



	@staticmethod
	def print_single(wred,ured,file_name="output",intro="",use_a1_paper=False,**opts):
		"""
		Print a single wiring diagrams into a file.

		Additionally, may contain ``file_name``, ``name``, ``new_page``, etc.
		as in the ``self.print_and_compile()`` function. Default values for all can be passed into ``opts``.

		See the documentation for ``self.print_and_compile()`` for input descriptions.
		"""
		DoubleBraidClusterA.print_multiple({"wred":wred, "ured":ured}, file_name=file_name, use_a1_paper=use_a1_paper, **opts)
