import os, re, random
import subprocess as sp

from WeylTools import *
from RQuiver import *

class DoubleBraidClusterGen:
	"""A class containing and computing double braid quivers in arbitrary type.
	See [GLSBS22]=[arXiv:2210.04778] and [GLSB23]=[arXiv:2301.07268]
	Code by Pavel Galashin, posted 01/16/23.

	DISCLAIMER:
	This code was written over the course of 4 years. It is messy and redundant and possibly incorrect. Use at your own risk.
	"""

	def __init__(self,letter=None, rank=None, typ=None, wred=None, ured="w0", speed_up=False,
				 u=None,n=None,lazy=False, debug=False, reverse_u=True, compute_labels=True, compute_quiver=True,
				 ignore_hollow_contributions=False):
		"""
			INPUT:

			 - ``wred`` -- double braid word in the alphabet 1,..,n-1,-1,...,-n+1.

			 - ``letter``, ``rank``, ``typ`` -- for e.g. D5, we specify either ``letter="D"``, ``rank=5`` or ``typ=["D",5]``

			 - ``ured`` -- either a reduced word (on positive letters), or "w0" or "id". Must be less than or equal the Demazure product of ``wred``.

			 - ``lazy``: if ``False``, will automatically run ``self.compute_data``, computing the quivers and everything else.

			 - ``reverse_u``: if True, then ``ured`` is reversed, so that positive letters in ``ured`` form an honest subword of ``wred``

			 - ``ignore_hollow_contributions`` -- whether to ignore the half-arrow/form contributions from hollow crossings
			 
			 - ``speed_up`` -- whether to use various heuristics to make the code run faster
		"""
		WT=WeylTools(letter=letter,rank=rank,typ=typ,use_dual=True)
		self.W,self.s,self.alpha,self.id,self.omega,self.w0,self.I, self.WT, self.rank,self.letter,self.typ= \
			WT.W,WT.s,WT.alpha,WT.id,WT.omega,WT.w0,WT.I, WT, WT.rank,WT.letter,WT.typ
		self.speed_up=speed_up
		wred_abs = [abs(j) for j in wred]
		self.id_perm=self.id
		self.n=self.rank
		if u==None:
			if ured == "w0":
				u = self.w0
			elif ured == "id":
				u = self.id_perm
			else:
				if reverse_u: ured=list(reversed(ured))
				u = WT.from_red(ured)
				if u.length() != len(ured): raise Exception("u is not reduced! ured=%s, len=%s"%(ured,u.length()))
		else:
			ured=WT.get_red(u)
		self.u=u
		self.ured=ured
		self.wlen=self.m=len(wred)
		self.wred=wred
		self.wred_abs=wred_abs
		self.w0_inv=WT.w0_involution_dic()
		self.cartan=WT.get_cartan_dic()
		self.symmetrizer_cartan=WT.W.cartan_type().symmetrizer()
		self.ignore_hollow_contributions=ignore_hollow_contributions
		try:
			NON_TRIVIAL_MUTATIONS
		except NameError:
			NON_TRIVIAL_MUTATIONS = []
		self.gamma_cache={}
		self.printed_gamma_cache={}
		self.max_debug_depth=oo
		if not lazy: self.compute_data(debug=debug,compute_labels=compute_labels, compute_quiver=compute_quiver)
			
	

	def compute_data(self,debug=False, compute_labels=True, compute_quiver=True):
		"""Compute the quiver and all the accompanying data."""
		wred,u,wred_abs,n,WT,w0,id_perm,wlen,typ,I,W= \
			self.wred,self.u,self.wred_abs,self.n,self.WT,self.w0,self.id_perm,self.wlen, self.typ, self.I, self.W
		if debug: print("%s computing data: u=%s, wred=%s"%(typ,WT.get_red(u),wred))
		nsymb="%"+str(3*n-3)+"s"
		m = wlen
		u_prev = u
		hollow = []
		solid = []
		uk={m:u_prev}
		for i in reversed(range(m)):
			wi, wia = wred[i], abs(wred[i])
			if wi < 0:
				u_next = u_prev * WT.s[wia]
			else:
				u_next = WT.s[wia] * u_prev
			if u_next.length() < u_prev.length():
				hollow.append(i)
				u_prev = u_next
			else:
				solid.append(i)
			uk[i]=u_prev
		hollow, solid = list(reversed(hollow)), list(reversed(solid))
		if debug: print("hollow+1=%s\nsolid+1=%s" % ([j + 1 for j in hollow], [j + 1 for j in solid]))
		if u_prev != id_perm:
			print("u_(0)=%s" % self.WT.get_red(u_prev))
			print("no subword for u!")

		current_left={i:-1 for i in I}
		left_neigh={}
		right_neigh={}
		for i in range(wlen):
			wi=abs(wred[i])
			left_neigh[i]=current_left[wi]
			right_neigh[current_left[wi]]=i
			current_left[wi]=i
		for wi in range(1,n):
			if current_left[wi]==-1: continue
			right_neigh[current_left[wi]]=oo

		self.hollow, self.solid, self.uk, \
			self.left_neigh, self.right_neigh = \
			hollow, solid, uk,\
			 left_neigh, right_neigh
		if compute_labels: self.compute_grid_labels(debug=debug)
		if compute_quiver: self.compute_quiver(debug=debug)
		if debug: print("%s computed double data! wred=%s, \nhollow+1=%s; solid+1=%s"%
						(typ,wred,[j+1 for j in hollow],[j+1 for j in solid]))

	def get_aps(self, disturb_index=-1):
		"""  Find almost positive subexpression for ``u``.
		 ``disturb_index`` tells us where the length goes down. """
		u_prev, wred, m, WT = self.u, self.wred, self.m, self.WT
		hollow = []
		solid = []
		uk = {m: u_prev}
		for i in reversed(range(m)):
			wi, wia = wred[i], abs(wred[i])
			if wi < 0:
				u_next = u_prev * WT.s[wia]
			else:
				u_next = WT.s[wia] * u_prev
			if (u_next.length() < u_prev.length()) or (disturb_index == i and (u_next.length() > u_prev.length())):
				hollow.append(i)
				u_prev = u_next
			else:
				solid.append(i)
			uk[i] = u_prev
		hollow, solid = list(reversed(hollow)), list(reversed(solid))
		return hollow, solid, uk

	
	def get_aps_akb(self, disturb_index):
		"""Find an APS/PDS pair in the form PDS*w0=a*k*b, APS*w0=a*b. See Section 7 of [GLSB23].
		 ``disturb_index`` tells us where the length goes down (must be solid).
		"""
		u_prev, wred, m, WT, w0_inv = self.u, self.wred, self.m, self.WT, self.w0_inv
		u_prev=u_prev*WT.w0
		ik=disturb_index
		for i in reversed(range(ik+1,m)):
			wi, wia = wred[i], abs(wred[i])
			if wi < 0:
				u_next = u_prev * WT.s[w0_inv[wia]]
			else:
				u_next = WT.s[wia] * u_prev
			if (u_next.length() > u_prev.length()):
				u_prev = u_next

		wik,wika=wred[ik],abs(wred[ik])
		if wik>0:
			a=ap=self.id
			b=u_prev=WT.s[wika]*u_prev
			k=wika
		else:
			b=self.id
			ap=self.id
			a=u_prev=u_prev*WT.s[w0_inv[wika]]
			k=w0_inv[wika]
		akb_dic={ik:(a,k,b)}
		for i in reversed(range(ik)):
			wi, wia = wred[i], abs(wred[i])
			a_next=a; b_next=b; ap_next=ap
			if wi < 0:
				b_next = b * WT.s[w0_inv[wia]]
				u_next = u_prev * WT.s[w0_inv[wia]]
			else:
				a_next = WT.s[wia] * a
				u_next = WT.s[wia] * u_prev
				if i in self.solid:
					ap_next = WT.s[wia] * ap
			if (a_next.length() > a.length()) or (b_next.length() > b.length()):
				a=a_next
				b=b_next
			akb_dic[i]=(a,k,b)
		return akb_dic


	def append_vu(self,Tv,Tu,also_return_vu,gamma,check_for_v_equal_to_u=False):
		"""If ``also_return_vu`` then return ``(gamma,Tv,Tu)``, otherwise return ``gamma``"""
		if check_for_v_equal_to_u and (Tv==Tu) and (gamma!=0*gamma):
			raise Exception("problem! Tv=Tu=%s, gammaR=%s" % (self.WT.get_red(Tv),self.WT.simple_roots_expansion_list(gamma)))
		if also_return_vu: return (gamma,Tv,Tu)
		else: return gamma

		
	def get_gamma(self, a, k, b, use_cache=True, debug=False, depth=0, also_return_vu=False):
		"""Algorithm from Section 7.
		Assume that a,b in W are reduced, 
		k in I such that a*s_k and s_k*b are reduced. 
		set v=a*k*b=PDS, u=a*b=APS. Return the ``\\gamma^+(a,k,b)`` vector.
		"""
		WT=self.WT
		if debug and depth>self.max_debug_depth: debug=False
		if isinstance(a,list):
			ared,bred=a,b
			a,b=WT.from_red(ared),WT.from_red(bred)
		else:
			ared,bred=WT.get_red(a),WT.get_red(b)
		Tv=WT.demazure_product(ared+[k]+bred)
		Tu=WT.demazure_product(ared+bred)
		if use_cache and ((a,k,b) in self.gamma_cache):
			return self.append_vu(Tv,Tu,also_return_vu,self.gamma_cache[a,k,b])
		if debug: print("ared,k,bred=",ared,k,bred)
		if Tv==Tu:
			zero=WT.alpha[1]*0
			self.gamma_cache[a,k,b]=zero
			if debug: print("[depth=%d] Tu=Tv, returning 0"%(depth))
			return self.append_vu(Tv,Tu,also_return_vu,zero)
		vu_equal="" if Tv!=Tu else "[Tv=Tu]"
		if debug: print("[depth=%d] %sgamma-new for a=%s, k=%s, b=%s"%(depth,vu_equal,ared,k,bred))
		if len(bred)==0:
			gammaR=a.action(self.alpha[k])
			self.gamma_cache[a,k,b]=gammaR
			if debug: print("[depth=%d] %sb=id, returning %s" % (depth,vu_equal,WT.simple_roots_expansion_list(gammaR)))
			return self.append_vu(Tv,Tu,also_return_vu,gammaR) #gammaR=a*alpha_k
		if a==WT.w0:
			zero=WT.alpha[1]*0
			self.gamma_cache[a,k,b]=zero
			if debug: print("[depth=%d] %sa=w_0, returning 0"%(depth,vu_equal))
			return self.append_vu(Tv,Tu,also_return_vu,zero)
		if Tv!=WT.w0: #Case 1a -- improved
			z=WT.w0*(~Tv)
			zi=~z
			gammaR_plus=self.get_gamma(WT.get_red(z) + ared, k, bred, depth=depth + 1, debug=debug, use_cache=use_cache)
			gammaR=zi.action(gammaR_plus)
			self.gamma_cache[a,k,b]=gammaR
			if debug: print("[depth=%d] %sCase 1a-impr, returning %s" % (depth,vu_equal,WT.simple_roots_expansion_list(gammaR)))
			return self.append_vu(Tv,Tu,also_return_vu,gammaR)
			
		for i in self.I:
			if (self.s[i]*a).length()>len(ared): break
		s=self.s[i]
		ap=s*a
		j=bred[-1]
		t=self.s[j]
		bp=b*t
		apred,bpred=[i]+ared,bred[:-1]
		Tv_plus=WT.demazure_product(apred+[k]+bred)
		Tv_minus=WT.demazure_product(ared+[k]+bpred)
		Tv_prime=WT.demazure_product(apred+[k]+bpred)

		if Tv_plus.length()>Tv.length(): #Case (1)
			gammaR_plus=self.get_gamma(apred, k, bred, depth=depth + 1, debug=debug, use_cache=use_cache)
			gammaR=s.action(gammaR_plus)
			self.gamma_cache[a,k,b]=gammaR
			if debug: print("[depth=%d] %sCase 1a, returning %s" % (depth,vu_equal,WT.simple_roots_expansion_list(gammaR)))
			return self.append_vu(Tv,Tu,also_return_vu,gammaR)
		if Tv.length()>Tv_minus.length(): #Case (2)
			gammaR_minus=self.get_gamma(ared, k, bpred, depth=depth + 1, debug=debug, use_cache=use_cache)
			gammaR=gammaR_minus
			if debug: print("[depth=%d] %sCase 1b, returning %s" % (depth,vu_equal,WT.simple_roots_expansion_list(gammaR)))
			self.gamma_cache[a,k,b]=gammaR
			return self.append_vu(Tv,Tu,also_return_vu,gammaR)
		if not (Tv==Tv_plus==Tv_minus==Tv_prime):
			raise Exception("ared=%s; k=%s; bred=%s; not (Tv==Tv_plus==Tv_minus==Tv_prime)"%(ared,k,bred))
		alpha=self.alpha[i]
		beta=-Tv.action(self.alpha[j])
		if debug: print("[depth=%d]i=%d, j=%d, alpha=%s; beta=%s; computing gammaR_plus/minus" % (
			depth, i, j, WT.simple_roots_expansion_dict(alpha), WT.simple_roots_expansion_dict(beta)))
		gammaR_plus=self.get_gamma(apred, k, bred, depth=depth + 1, debug=debug, use_cache=use_cache)
		gammaR_minus=self.get_gamma(ared, k, bpred, depth=depth + 1, debug=debug, use_cache=use_cache)
		if alpha!=beta: #Case (3a)
			found_nonzero=False
			den=None
			for ell in self.I:
				if ell==i: continue
				den=WT.comega[ell].scalar(beta)
				if den!=0: found_nonzero=True;  break
			if not found_nonzero: raise Exception("""[depth=%d] ared=%s; k=%s; bred=%s; i=%d, j=%d, 
g^+=%s,  
g^-=%s, 
alpha=%s
beta =%s
couldn't find nonzero ell!""" % (depth, ared, k, bred, i, j, WT.simple_roots_expansion_list(gammaR_plus), WT.simple_roots_expansion_list(gammaR_minus), WT.simple_roots_expansion_list(alpha), WT.simple_roots_expansion_list(beta)))
			y=WT.comega[ell].scalar(gammaR_plus-gammaR_minus)/den
			gammaR=gammaR_minus+y*beta
			if debug: print("[depth=%d] %sCase 2a, returning %s" % (depth,vu_equal,WT.simple_roots_expansion_list(gammaR)))
			self.gamma_cache[a,k,b]=gammaR
			return self.append_vu(Tv,Tu,also_return_vu,gammaR)
		                #Case (3b)
		gammaR_prime=self.get_gamma(apred, k, bpred, depth=depth + 1, debug=debug)
		algp=WT.simple_roots_expansion_dict(gammaR_prime)
		min1=WT.comega[i].scalar(gammaR_plus+gammaR_minus)
		min2=sum([-self.cartan[i,k]*algp[k] for k in self.I if k!=i])
		RHS=min(min1,min2)-algp[i]
		alg=copy(algp)
		alg[i]=RHS
		gammaR=sum([alg[k]*self.alpha[k] for k in self.I])
		self.gamma_cache[a,k,b]=gammaR
		if debug: print("[depth=%d] %sCase 2b, returning %s" % (depth,vu_equal,WT.simple_roots_expansion_list(gammaR)))
		return self.append_vu(Tv,Tu,also_return_vu,gammaR)


	def compute_grid_labels(self, debug=False):
		""" Compute how grid minors factor into irreducibles. Computes (``i=1,2``):

		 - ``gamma``: the weight for each cluster var at each vertical section

		 - ``F_lab_i``: factorization for each grid minor (given by two coords ``i,h``) into irreducibles  """
		I,m,WT,w0=self.I,self.m,self.WT,self.w0
		akbs={}
		gammavus={}
		for r in self.solid:
			akb_dic = self.get_aps_akb(disturb_index=r)
			akbs[r]=akb_dic
			gammavus[r]={}
			for i in range(r+1):
				if self.speed_up and not (i==0 or i in self.solid or i-1 in self.solid): continue #i=0 for frozens
				a,k,b=akb_dic[i]
				# gammavus[r][i]=self.get_gamma(a,k,b,also_return_vu=True)
				gammavus[r][i]=self.get_gamma(a, k, b, also_return_vu=True)

		F_lab_1 = {(i, h): [] for i in range(m + 1) for h in I}
		F_lab_2 = {(i, h): [] for i in range(m + 1) for h in I}

		F_lab_1_dic = {(i, h): {} for i in range(m + 1) for h in I}
		F_lab_2_dic = {(i, h): {} for i in range(m + 1) for h in I}

		for r in self.solid:
			for i in range(r+1):
				if self.speed_up and not (i==0 or i in self.solid or i-1 in self.solid): continue #i=0 for frozens
				gR,v,u=gammavus[r][i] #v=PDS*w0
				alvg=WT.simple_roots_expansion_dict(gR)
				gL=-(~v).action(gR)
				alg=WT.simple_roots_expansion_dict(-w0.action(gL))
				for h in I:
					if not (alg[h]>=0 and alvg[h]>=0 and alg[h] in ZZ and alvg[h] in ZZ):
						a,k,b=akbs[r][i]
						print("wred=%s; ured=%s"%(self.wred,self.ured))
						print("PROBLEM: a=%s, k=%d, b=%s"%(WT.get_red(a),k,WT.get_red(b)))
						raise Exception("r=%d; i=%d; h=%s, gamma=%s, v.gamma=%s -- problem!\n%s%s%s%s"%(r,i,h,alg,alvg,
										alg[h]>=0, alvg[h]>=0, alg[h] in ZZ,  alvg[h] in ZZ))	  
					if alvg[h]>0:
						F_lab_1_dic[i,h][r]=alvg[h]
						F_lab_1[i,h]+=[r]*ZZ(alvg[h])
					if alg[h]>0:
						F_lab_2_dic[i,h][r]=alg[h]
						F_lab_2[i,h]+=[r]*ZZ(alg[h])
		self.F_lab_1, self.F_lab_2, self.F_lab_1_dic, self.F_lab_2_dic, self.gammavus, self.akbs = \
			F_lab_1, F_lab_2, F_lab_1_dic, F_lab_2_dic, gammavus, akbs

	def print_grid_labels(self):
		"""Assuming ``self.compute_grid_labels()`` has been computed  """
		F_lab_1, F_lab_2, I, m = self.F_lab_1, self.F_lab_2, self.I, self.wlen
		for h in reversed(I):
			labels = [[j + 1 for j in F_lab_1[i, h]] for i in range(m + 1)]
			print("row=%d; labels top=%s" % (h, "|".join(["%8s" % (lab) for lab in labels])))
		print()
		for h in reversed(I):
			labels = [[j + 1 for j in F_lab_2[i, h]] for i in range(m + 1)]
			print("row=%d; labels bot=%s" % (h, "|".join(["%8s" % (lab) for lab in labels])))

	def compute_quiver(self,debug=False,r1_debug=-1,r2_debug=-1):
		""" Compute the quiver:

		 - ``self.frozen_list``: list of frozen vertices

	 	- ``self.quiver``: the quiver as a dictionary ``(i,j)``:[nonnegative] number of arrows ``i->j``
		 """
		F_lab_1, F_lab_2, I, m, wred, cartan = self.F_lab_1, self.F_lab_2, self.I, self.wlen, self.wred, self.cartan
		frozens={}
		for h in I:
			for r in F_lab_1[0,h]+F_lab_2[0,h]:
				frozens[r]=True
		frozen_list=self.frozen_list=sorted(list(frozens))
		region_pairs=[]
		half_arrows={}
		i_list=self.solid if self.ignore_hollow_contributions else range(m)
		d=self.symmetrizer_cartan
		self.d_sym={i:d[abs(wred[i])] for i in self.solid}
		for i in i_list:
			wi,wia=wred[i],abs(wred[i])
			di=d[wia]
			if wi>0:
				chL=F_lab_1[i,wia]
				chR=F_lab_1[i+1,wia]
				region_pairs+=[(chR,chL,2*di)]
				for j in I:
					if j==wia: continue
					aij,aji=ZZ(cartan[wia,j]), ZZ(cartan[j,wia])
					if aij==0: continue
					if not (aij<0): raise Exception("wia=%d, j=%d, aij=%s -- unexpected!"%(wia,j,aij))
					if F_lab_1[i,j] != F_lab_1[i + 1, j]:
						print("%s wred,ured=%s, %s"%(self.typ,self.wred,self.ured))
						raise Exception(
							"top_1 labels not same! i=%d, %s != %s" % (i, F_lab_1[i,j], F_lab_1[i + 1, j]))
						# print(
						# 	"top_1 labels not same! i=%d, %s != %s" % (i, F_lab_1[i,j], F_lab_1[i + 1, j])) #TODO
					chY=F_lab_1[i,j]
					contrib=di*abs(aij)
					region_pairs+=[(chL,chY,contrib),(chY,chR,contrib)]
					# region_pairs+=[(chL,chY)]*abs(aij)
					# region_pairs+=[(chY,chR)]*abs(aji)
			else:
				chL=F_lab_2[i,wia]
				chR=F_lab_2[i+1,wia]
				region_pairs+=[(chL,chR,2*di)]
				for j in I:
					if j==wia: continue
					aij,aji=ZZ(cartan[wia,j]), ZZ(cartan[j,wia])
					if aij==0: continue
					if not (aij<0): raise Exception("wia=%d, j=%d, aij=%s -- unexpected!"%(wia,j,aij))
					if F_lab_2[i,j] != F_lab_2[i + 1, j]:
						print("%s wred,ured=%s, %s"%(self.typ,self.wred,self.ured))
						# print(
						# 	"top_2 labels not same! i=%d, %s != %s" % (i, F_lab_2[i,j], F_lab_2[i + 1, j]))
						raise Exception(
							"top_2 labels not same! i=%d, %s != %s" % (i, F_lab_2[i,j], F_lab_2[i + 1, j]))
					chY=F_lab_2[i,j]
					contrib=di*abs(aij)
					region_pairs+=[(chR,chY,contrib),(chY,chL,contrib)]
					# if (r1_debug in chR+chY+chL and r2_debug in chR+chY+chL):
						
					# region_pairs+=[(chR,chY)]*abs(aij)
					# region_pairs+=[(chY,chL)]*abs(aji)
		for A,B,contrib in region_pairs:
			for r1 in A:
				for r2 in B:
					if sorted([r1,r2])==sorted([r1_debug,r2_debug]):
						print("[debug] r1=%d, r2=%d, A=%s, B=%s"%(r1,r2,A,B))
					if (r1,r2) not in half_arrows:
						half_arrows[r1,r2]=half_arrows[r2,r1]=0
					half_arrows[r1,r2]+=contrib
					half_arrows[r2,r1]-=contrib
		quiver={}
		for r1,r2 in half_arrows:
			narr=half_arrows[r1,r2]
			if narr<=0: continue
			if (r1 in frozen_list) and (r2 in frozen_list): continue
			# if narr>2: print("wred,ured=%s,%s\nLOTS OF ARROWS! r1+1=%d, r2+1=%d, multiplicity=%d"%(
			# 	wred, self.WT.get_red(self.u),r1+1,r2+1,narr//2)) #example with multi arrows:
			# # wred, ured = [5, -2, 2, -3, 2, -4, 6, 4, -4, 5, 1, 1, 4, 4, 4, 3, 2, 1, 5, 4, 3, 3, 2, 5, 4, 6, 5, 4, 3, 2,
			# # 			  1, 1, 2, 4, 5, 6, 5, 5, 5], [6, 4, 5, 6, 3, 4, 5, 2, 3, 4, 5, 6, 1, 2, 3, 4, 5, 6]  # mult arr
			# # if narr>2: print("LOTS OF ARROWS! r1=%d, r2=%d, narr=%d"%(r1+1,r2+1,narr))
			if narr%2!=0:
				print("typ=%s\nwred=%s\nured=%s\nsolid+1=%s\nhollow+1=%s"%(self.typ,self.wred,self.ured,\
								    [j+1 for j in self.solid],[j+1 for j in self.hollow]))
				raise Exception("odd number of half-arrows! r1=%d, r2=%d, narr=%d"%(r1+1,r2+1,narr))
			quiver[(r1,r2)]=(narr//2)
		self.quiver=quiver
		

	def get_RQuiver(self,debug=False,index_add=0,ignore_frozens=False):
		"""Return an ``RQuiver`` object whose B-matrix is skew-symmetrizable """
		solid, frozen_list, quiver, d_sym = self.solid, self.frozen_list, self.quiver, self.d_sym
		remap={}
		remap_i={}
		cnt=0
		d_sym_remapped={} # keys=range(n)
		for i in solid:
			if i not in frozen_list:
				remap[i+index_add]=cnt
				remap_i[cnt]=i+index_add
				d_sym_remapped[cnt]=d_sym[i]
				cnt+=1
		for i in reversed(frozen_list):
			remap[i+index_add]=cnt
			remap_i[cnt]=i+index_add
			d_sym_remapped[cnt]=d_sym[i]
			cnt+=1
		if debug: print("solid=%s, frozen_list=%s, remap=%s"%(solid,frozen_list,remap))
		n_fro=m=len(frozen_list)
		n_mut=n=len(solid)-m
		B=Matrix(n+m,n)
		cartan,wred=self.cartan,self.wred

		for i,j in quiver:
			ri=remap[i+index_add]
			rj=remap[j+index_add]
			if rj<n: B[ri,rj]+=quiver[i,j]/d_sym[i]
			if ri<n: B[rj,ri]-=quiver[i,j]/d_sym[j]
		B.set_immutable()
		return RQuiver(B,labels=remap_i,ignore_frozens=ignore_frozens,d_sym=d_sym_remapped,check_skew_symm=True)



	@staticmethod
	def compile_latex(text, file_name="output", tex_dir="tex/", template_path="tex/double_template_gen.tex",
					  error_from=500, error_to=500, return_compil_output=False, clean_aux=True, use_a1_paper=False):
		"""Output ``text`` into a latex file ``file_name``.tex using ``template_dir/template_path`` template file and compile.

		INPUT:
		 - ``error_from`` and ``error_to`` determine neighborhood size of an error message to output when error occurs.

		 - ``template_path`` is path to template file. Must contain "%PASTE%" where we will paste our new text
		output file is found in ``tex_dir/`` + ``file_name``.tex
		"""
		f=open(template_path,"r");
		template=f.read();
		subbed=template.replace("%PASTE%",text);
		if use_a1_paper:
			subbed=subbed.replace("usepackage[margin=1in]{geometry}","usepackage[a1paper,margin=1in]{geometry}")
		o=open(tex_dir+file_name+".tex","w");
		o.write(subbed);
		o.close()
		command='latexmk -pdf -f -g -bibtex -synctex=1 -halt-on-error -jobname=%s%s %s%s.tex'%(tex_dir,file_name,tex_dir,file_name)

		pipe = sp.Popen( command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE )
		res = pipe.communicate()
		compil=res[0].decode("utf-8")

		if "error" in compil or "Error" in compil or pipe.returncode!=0:
			print("latex compilation error! return code=",pipe.returncode)
			for m in re.finditer('[Ee]rror', compil):
				print("*************************************\n")
				print(compil[max(0,m.start()-error_from):min(len(compil),m.end()+error_to)])
		# else: print("OK! file name="+file_name)
		# else: html("OK! click: <a href='%s'>%s</a>"%(tex_dir+file_name+".pdf",file_name+".pdf"))
		else:
			html("OK! <a href='%s'>%s</a>"%(tex_dir+file_name+".pdf",file_name+".pdf"))
		if clean_aux:
			for ext in ["*.aux","*.log","*.fdb_latexmk","*.fls","*.out","*.blg","*.nav","*.snm","*.toc","*.synctex.gz"]:
				command="rm "+tex_dir+ext
				pipe = sp.Popen( command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE )
				res = pipe.communicate()
		if return_compil_output: return compil


	def get_tikzpicture(self, omit_trail=True, index_add=0, right_aligned=True, debug=False, compute_matrices=False,
				 grid_label_scale=None, print_red_factorizations=True, print_blue_factorizations=True,
				 draw_quiver=True, only_mutable_quiver=False,F_shift=0, special=-1,
				 print_special_arrows=True):
		"""Outputs LaTeX source as a tikzpicture representing our wiring diagram.
		See the documentation for ``self.print_and_compile()`` for input descriptions."""
		wred = self.wred
		I, hollow, solid, left_neigh, right_neigh, wred, typ = \
			self.I, self.hollow, self.solid, self.left_neigh, self.right_neigh, self.wred, self.typ
		uk, w0, WT, wlen = self.uk, self.w0, self.WT, self.wlen
		F_lab_1,F_lab_2=self.F_lab_1,self.F_lab_2
		if omit_trail: wlen = max(solid) + 1
		m=wlen
		xscale = min(0.55, 0.55 * 12 / wlen)
		yscale = min(0.8, 0.8 * 5 / self.rank)
		if grid_label_scale==None:
			grid_label_scale=min(0.5,0.5*20/wlen)
		tpic = ""
		tpic += ("\\begin{tikzpicture}[xscale=%.2f,yscale=%.2f]\n") % (xscale, yscale)
		tpic += ("\\def\\xsh{%.2f}\n" % (0.5 / xscale * 0.4))
		tpic += ("\\def\\ysh{%.2f}\n" % (0.8 / yscale * 0.08))
		extra_rb = 1 if right_aligned else 2
		tpic += "\\def\\rbound{" + str(wlen * 2 + extra_rb) + "}\n"
		tpic += "\\def\\rboundOverTwo{" + str(wlen + extra_rb // 2) + "}\n"
		tpic += "\\def\\grid_label_scale{" + str(grid_label_scale) + "}\n"
		tpic += "\\wiresAB{" + str(self.rank+1) + "}\n"
		cartan=self.cartan
		dynkin_lw=1.5
		dynkin_lw_thin=0.5
		dynkin_bend=50
		n=self.rank+1
		for i in I:
			for j in I:
				if i>=j: continue
				aij=abs(cartan[i,j])
				aji=abs(cartan[j,i])
				if aij==0: continue
				if aij==aji==1:
					if abs(i-j)==1:
						tpic+="\\draw[line width=%.2fpt,purple] (LN%d)--(LN%d);\n"%(dynkin_lw,i,j)
						tpic+="\\draw[line width=%.2fpt,purple] (RN%d)--(RN%d);\n"%(dynkin_lw,i,j)
					else:
						tpic+="\\draw[line width=%.2fpt,purple] (LN%d) to[bend left=%.2f] (LN%d);\n"%\
						               (dynkin_lw,i,dynkin_bend,j)
						tpic+="\\draw[line width=%.2fpt,purple] (RN%d) to[bend right=%.2f] (RN%d);\n"%\
						               (dynkin_lw,i,dynkin_bend,j)
				else:
					narr=aij*aji
					if narr not in [2,3]: raise Exception("unexpected narr=%d"%(narr))
					style="->,double" if narr==2 else "->,triple"
					x,y=(i,j) if aij<aji else (j,i)
					if abs(i-j)==1:
						tpic+="\\draw[line width=%.2fpt,purple,%s] (LN%d)--(LN%d);\n"%(dynkin_lw_thin,style,x,y)
						tpic+="\\draw[line width=%.2fpt,purple,%s] (RN%d)--(RN%d);\n"%(dynkin_lw_thin,style,x,y)
					else:
						raise NotImplemented("i=%d, j=%d, aij=%d, aji=%d -- far apart!"%(i,j,aij,aji))
		tpic+="\\node[purple,scale=1,anchor=east] (LN) at (-1+\\lbound,%.2f) {$%s_{%d}$};\n"%(0.5+n/2,typ[0],typ[1])
		tpic+="\\node[purple,scale=1,anchor=west] (RN) at (1+\\rbound,%.2f) {$%s_{%d}$};\n"%(0.5+n/2,typ[0],typ[1])
				
		i_list = list(range(wlen)) if right_aligned else reversed(list(range(wlen)))
		for i in i_list:
			wi, wia = wred[i], abs(wred[i])
			h_coord = str(i)
			neigh = -2
			if right_aligned:
				if left_neigh[i] >= 0:
					lene = str(left_neigh[i])
					neigh = left_neigh[i]
				else:
					lene = "\\lboundOverTwo"
			else:
				if right_neigh[i] != oo:
					lene = str(right_neigh[i])
					neigh = right_neigh[i]
				else:
					lene = "\\rboundOverTwo"
			Fs_str_list = " "
			if i not in hollow:
				if wi > 0:
					tpic += "\\dotSEPfilledA{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
				else:
					tpic += "\\dotSEPfilledB{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
			else:
				if wi > 0:
					tpic += "\\xdotSEPfilledA{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
				else:
					tpic += "\\xdotSEPfilledB{%s}(%s,%s){%s}{%s}\n" % (
					Fs_str_list, h_coord, wia, lene, i + 1 + index_add)
		lab_1,lab_2=(F_lab_1,F_lab_2)
		sep=""
		extra_add=1
		max_lab_len=1
		sep_coef=0.5
		for i in range(m+1):
			for h in I:
				if print_red_factorizations:
					lab = sep.join([str(r+extra_add+index_add) for r in lab_1[i, h]])
					max_lab_len=max(max_lab_len,len(lab)+sep_coef*(len(lab_1[i,h])-1))
				if print_blue_factorizations:
					lab = sep.join([str(r+extra_add+index_add) for r in lab_2[i, h]])
					max_lab_len=max(max_lab_len,len(lab)+sep_coef*(len(lab_1[i,h])-1))
		lab_scale=min(grid_label_scale,grid_label_scale*6/max_lab_len)
		sep="\\!\\cdot\\!"
		for i in range(m+1):
			for h in I:
				if print_red_factorizations:
					lab = sep.join([str(r+extra_add+index_add) for r in lab_1[i, h]])
					if  (special in [r+1 for r in lab_1[i, h]]):
						tpic += "\\draw[red!40,fill=brown!40] (%.2f,%.2f) circle (0.25cm);\n" % (2*i-0.5, h+0.7)
					tpic += "\\node[red,scale=%.2f] (F1_%d_%d) at (%.2f,%.2f) {$%s $};\n" % \
						(lab_scale, i, h, 2*i-0.5, h+0.7, lab)
				if print_blue_factorizations:
					lab = sep.join([str(r+extra_add+index_add) for r in lab_2[i, h]])
					if  (special in [r+1 for r in lab_2[i, h]]):
						tpic += "\\draw[red!40,fill=brown!40] (%.2f,%.2f) circle (0.25cm);\n" % (2*i-0.5, h+0.3)
					tpic += "\\node[blue, scale=%.2f] (F1_%d_%d) at (%.2f,%.2f) {$%s $};\n" % \
						(lab_scale, i, h, 2*i-0.5, h+0.3, lab)
		mult_arrows=[]
		if draw_quiver:
			for r in self.solid:
				wi,wia=wred[r],abs(wred[r])
				if r in self.frozen_list:fill="black!15"
				else: fill="green!50"
				if wi>0: draw="red!70!black"; ysh="-\\shiftAB"
				else: draw="blue!70!black"; ysh="+\\shiftAB"

				tpic+="\\node[black,draw=%s,%s,fill=%s,rectangle,scale=%.2f] (F%d) at (%.2f,%.2f%s) {$%d$};\n"%\
					  (draw,draw, fill, grid_label_scale, r, 2*r+0.5-F_shift, wia+0.5,ysh, r+1+index_add)
			Q=self.get_RQuiver(index_add=index_add)
			B=Q.B
			d=self.d_sym
			quiver=self.quiver
			for x, y in quiver:
				if only_mutable_quiver and (x in self.frozen_list or y in self.frozen_list): continue
				if (quiver[x, y] % d[x] != 0) or (quiver[x, y] % d[y] != 0):
					raise Exception("x=%s, y=%s, quiver[x,y]=%s, d[x]=%s, d[y]=%s --- not divisible!"%(x,y,quiver[x,y],d[x],d[y]))
				# mult = quiver[x,y]
				print_mult=False
				# mult1, mult2 = quiver[x, y] / d[x], - quiver[x, y] / d[y]
				# if not (x in self.frozen_list or y in self.frozen_list):
				Bxy=Q.Bij_from_label(x+index_add,y+index_add)
				Byx=Q.Bij_from_label(y+index_add,x+index_add)
				if not (Bxy==1 and Byx==-1):
					mult="%d:%d"%(Bxy,Byx)
					print_mult=True
					mult_arrows.append(" $%d\\xrightarrow{%d:%d}%d$ "%(x+1+index_add,Bxy,Byx,y+1+index_add))
				xs,ys=str(x),str(y)
				if (x in self.frozen_list or y in self.frozen_list):
					dred = "drawedgeExp"
				else:
					dred = "drawedgeThick"
				if not print_mult:
					tpic+="\\%s{%s}{%s}{%s}\n"%(dred,xs,ys,self.get_color(x + index_add, y + index_add))
				else:
					dred+="Mult"
					tpic+="\\%s{%s}{%s}{%s}{%s}\n"%\
						(dred,xs,ys,self.get_color(x + index_add, y + index_add),str(mult))
		tpic += "\\end{tikzpicture}\n"
		if mult_arrows!=[]: tpic+="\nskew-symmetrizable arrows: %s\n"%(", ".join(mult_arrows))
		if print_special_arrows and special!=-1 and draw_quiver:
			spec_arrows_in=[]
			spec_arrows_out=[]
			spec_arrows_mult=[]
			for x,y in quiver:
				if quiver[x,y]==1:
					if x+1 == special: spec_arrows_out.append("%d\\to%d"%(x+1,y+1))
					if y+1 == special: spec_arrows_in.append("%d\\to%d"%(x+1,y+1))
				else:
					if special in [x+1,y+1]: spec_arrows_mult.append("%d\\to%d: %d"%(x+1,y+1,quiver[x,y]))
			tpic+="\nspecial=$%d$: arrows in=$[%s]$; arrows out=$[%s]$"%(special,",".join(spec_arrows_in),",".join(spec_arrows_out))
			if len(spec_arrows_mult)>0: tpic+=";  MULT ARROWS=$[%s]$"%(",".join(spec_arrows_mult))
		return tpic + "\n\n\n"

	def get_color(self,x,y):
		return str((max(x,y)*71)%100)


	@staticmethod
	def print_and_compile(letter=None, rank=None, typ=None, wred=None, ured="w0", file_name="output", right_aligned=True,
					 name="", comment="", new_page=False, return_tex=False, omit_trail=False, debug=False, grid_label_scale=None,
					 print_uw=True, index_add=0, intro="", speed_up=False, ignore_hollow_contributions=False,
					 print_red_factorizations=True, print_blue_factorizations=True, draw_quiver=True, special=-1, reverse_u=True,
					 print_cartan=False):
		"""Print everything to ``file_name``.tex. Compile it if ``return_tex=False``, otherwise, return LaTeX source.

		INPUT:
		 - ``wred/ured/letter/rank/typ`` -- same as for this class.
		 - ``file_name`` -- name of the output file
		 - ``name`` -- title of the section
		 - ``comment`` -- text above the wiring diagram picture
		 - ``right_aligned`` -- whether to center labels inside wiring diagram regions or make them right-aligned
		 - ``new_page`` -- whether to add \\newpage before
		 - ``omit_trail`` -- if ``True``, hollow crossings at the right end of the wiring diagram are omitted
		 - ``print_uw`` -- whether to print the reduced words for ``u`` and ``w``
		 - ``index_add`` -- increase/decrease all indices by some integer. Useful when chopping off letters from the beginning of ``wred``.
		 - ``compute_matrices`` -- whether to compute the ``X_c,Y_c,Z_c`` matrices
		 -  ``intro`` -- LaTeX text to be added at the beginning
		 - ``speed_up`` -- whether to use various heuristics to make the code run faster
		 - ``ignore_hollow_contributions`` -- whether to ignore the half-arrow/form contributions from hollow crossings
		 - ``print_red/blue_factorizations`` -- whether to print how each red/blue grid minor factors into cluster vars
		 - ``draw_quiver`` -- whether to draw the arrows
		 - ``use_set_labels`` -- whether to label the regions by the corresponding sets of rows/columns
		 - ``special`` -- index of a cluster variable that will be highlighted in the wiring diagram
		 - ``reverse_u`` -- if true, ``ured`` is reversed
		 - ``print_cartan`` -- whether to print the Cartan matrix to LaTeX. Useful to nail conventions for multiply-laced types.
		 """
		tpics = ""
		if new_page: tpics += "\\newpage"
		ld = DoubleBraidClusterGen(letter=letter, rank=rank, typ=typ, wred=wred, ured=ured, reverse_u=reverse_u,
								   compute_quiver=draw_quiver, speed_up=speed_up,
								   ignore_hollow_contributions=ignore_hollow_contributions)
		if name != None: tpics += "\\section{%s}\n\n" % (name)
		wred_strs = []
		m = len(wred)
		hollow, solid = ld.hollow, ld.solid
		if print_uw:
			for i in range(m):
				wi, wia = wred[i], abs(wred[i])
				if wi > 0:
					color = "red"
				else:
					color = "blue"
				letter = "\\textcolor{%s}{%s}" % (color, wi)
				if i in ld.hollow: letter = "\\boxed{%s}" % letter
				wred_strs.append(letter)
			wred_str = "[%s]" % (",".join(wred_strs))
			tpics += "\\begin{align*} u&=\\textcolor{black}{%s}\\\\ w&=%s \\end{align*} \n\n" % (list(reversed(ld.ured)), wred_str)
		if comment != "": tpics += "\n %s \n\n" % comment
		tpics += ld.get_tikzpicture(right_aligned=right_aligned, grid_label_scale=grid_label_scale,
							 omit_trail=omit_trail, debug=debug, 
							 index_add=index_add, print_red_factorizations=print_red_factorizations,
							 print_blue_factorizations=print_blue_factorizations,draw_quiver=draw_quiver, special=special)
		if print_cartan:
			cmat=Matrix(ld.cartan)
			r=ld.rank
			cmat=cmat.matrix_from_rows_and_columns(range(1,r+1),range(1,r+1))
			tpics+="\n\nCartan matrix $a_{ij}=\\<\\alpha^\\vee_i,\\alpha_j\\>$: \n\n $%s$"%(latex(cmat))
		if return_tex:
			return tpics
		else:
			DoubleBraidClusterGen.compile_latex(tpics, file_name=file_name)


	def print_to_file(self,file_name="output",**opts):
		tpic=self.get_tikzpicture(opts)
		DoubleBraidClusterGen.compile_latex(tpic, file_name=file_name)

	@staticmethod
	def print_multiple(datas,file_name="output",intro="",use_a1_paper=False,**opts):
		"""Print several wiring diagrams into a single file.

		INPUT:
		 - ``datas``=a list of dictionaries of quivers to print.
		each item of ``datas`` has to contain a "ured" and a "wred" parameters.
		Additionally, it may contain ``file_name``, ``name``, ``new_page``, etc.
		as in the ``self.print_and_compile()`` function. Default values for all can be passed into ``opts``.

		See the documentation for ``self.print_and_compile()`` for input descriptions."""
		global prev_quiver_local
		prev_quiver_local = None
		tpics=intro
		if not isinstance(datas,list):
			if isinstance(datas,dict): datas=[datas]
			else: raise Exception("expected a list, got: %s"%datas)
		# print("intro=",intro)
		for dat in datas:
			dat2=opts.copy()
			# print(dat2,dat)
			dat2.update(dat)
			dat2["return_tex"]=True
			if "intro" in dat2: dat2.pop("intro")
			tpics+=DoubleBraidClusterGen.print_and_compile(**dat2)
		DoubleBraidClusterGen.compile_latex(tpics, file_name=file_name,use_a1_paper=use_a1_paper)



	@staticmethod
	def print_single(letter=None, rank=None, typ=None, wred=None, ured="w0",file_name="output",intro="",use_a1_paper=False,**opts):
		"""
		Print a single wiring diagrams into a file.

		Additionally, may contain ``file_name``, ``name``, ``new_page``, etc.
		as in the ``self.print_and_compile()`` function. Default values for all can be passed into ``opts``.

		See the documentation for ``self.print_and_compile()`` for input descriptions.
		"""
		DoubleBraidClusterGen.print_multiple({"letter":letter, "rank":rank, "typ":typ,
									  "wred":wred,"ured":ured}, file_name=file_name, intro=intro, use_a1_paper=use_a1_paper, **opts)
