# load("MatrixTools.py","RQuiver.py")

import os, re, random;
import subprocess as sp

from MatrixTools import *
from RQuiver import *


class LecDataBraid:
	"""a class containing and computing various things in relation to Gracie's quivers *for braid varieties*. Namely:
	``n``,``v``,``wred``,``wlen`` -- clear (``w`` is not a permutation anymore!)
	``used``, ``skipped``: indices used for subexpression for v inside wred. (``skipped``=fat dots)
	``frozen_list``: which indices are frozen, numbered from 0
	``Fs_used_list``: for each index (num from 0), record its factors into irreducibles (num from 1)
	``JCj_region``: for each skipped index (num from 0), record the indices (num from 1) where it appears as a factor
	``quiver``: the quiver as a dic (i,j):number of arrows i->j, only positive
	``left_neigh``: for each index i, record the index j to the left of it such that wred[i]=wred[j]
	``expected_arrows``: BFZ arrows
	``v_sets_right``: source-labeled bottom v-strand labels of the rightmost regions of the wiring diagram"""
	def __init__(self,vred,wred,n=None,lazy=False,debug=False):
		if n==None:
			n=max(vred+wred)+1;
		self.mt=MatrixToolsOld(n=n);
		self.n=n;
		self.vred=vred;
		self.wred=wred;
		self.wlen=len(wred)
		if not lazy: self.compute_data(debug=debug);
			
	
	
	def __get_v_labels(self,disturb_index=None,debug=False):
		"""for each (i,h), this is the set of v-strands below, source labeled, of the region to the left of i-th crossing, at height h
		``disturb_index`` tells us at which index do we take the maximum instead of the minimum"""
		if debug: print("computing v_labels for disturb index=%s"%disturb_index)
		n,wred,v_sets_right,wlen=self.n, self.wred, self.v_sets_right,self.wlen
		v_labels={}
		for i in reversed(range(wlen+1)):
			if i==wlen:
				for h in range(n+1):
					v_labels[i,h]=v_sets_right[h]
				continue;
			for h in range(n+1):
				v_labels[i,h]=v_labels[i+1,h]
			h=wred[i]
			A,Aij=v_labels[i+1,h-1],v_labels[i+1,h+1]
			ij=set(Aij).difference(set(A))
			if i!=disturb_index:
				v_labels[i,h]=sorted(A+[min(ij)])
			else:
				v_labels[i,h]=sorted(A+[max(ij)])
			if debug: print("v_label[%d,%d]=%s"%(i,h,v_labels[i,h]))
		return v_labels


	def compute_data(self,debug=False):
		vred,wred,n,mt=self.vred,self.wred,self.n,self.mt;
		v=mt.from_red(vred);
		if debug: print("computing data: vred=%s, v=%s, wred=%s, n=%d"%(vred,v,wred,n));
		wlen=len(wred)
		if len(vred)!=v.length(): raise Exception("v not reduced! vred=%s; v.length=%d"%(vred,v.length()))
		nsymb="%"+str(3*n-3)+"s";
		used,BS=mt.get_positive_subexpr(v=v,wred=wred)
		skipped=sorted(list(set(range(wlen)) - set(used)));
		if debug: print("used=",used,";   skipped=",skipped,"\n")

		expected_arrows={}
		current_left=[-1]*(n+1);
		left_neigh={};
		right_neigh={}

		for i in range(wlen):
			wi=wred[i];
			if current_left[wi]>=0:
				expected_arrows[(i,current_left[wi])]=True;
			left_neigh[i]=current_left[wi];
			right_neigh[current_left[wi]]=i;
			current_left[wi]=i;
		for i in range(wlen):
			if i not in right_neigh:
				right_neigh[i]=wlen+1;
		current_left=[-1]*(n+1);
		for i in range(wlen):
			wi=wred[i];
			if current_left[wi+1]>=0 and current_left[wi+1]>=current_left[wi]:
				expected_arrows[(i,right_neigh[current_left[wi+1]])]=True;
			if current_left[wi-1]>=0 and current_left[wi-1]>=current_left[wi]:
				expected_arrows[(i,right_neigh[current_left[wi-1]])]=True;
			current_left[wi]=i;
		if debug: print("expected arrows: ",list(expected_arrows))
		if debug: mt.print_positive_subexpr(v=v,wred=wred)
		vi=~v;
		v_sets_right=[[]]
		for h in range(n):
			v_sets_right.append(sorted(v_sets_right[h]+[vi(h+1)]));
		if debug: print("v_sets_right=",v_sets_right);
		self.v_sets_right=v_sets_right
		
		v_labels=self.__get_v_labels(disturb_index=None,debug=debug)
		v_labels_dist={i: self.__get_v_labels(disturb_index=i,debug=False) for i in skipped};

		JCj_region={}
		Fs_used_list=[[] for _ in range(wlen)]
		frozen_list=[]

		for i in skipped:
			JCj_region[i]=[]
			vld=v_labels_dist[i]
			for j in range(i+1):
				h=wred[j]
				if v_labels[j,h]!=vld[j,h]:
					JCj_region[i].append(j+1)
					Fs_used_list[j].append(i+1)
			is_frozen=False
			for h in range(1,n):
				if v_labels[0,h]!=vld[0,h]: is_frozen=True; break
			if is_frozen: frozen_list.append(i);
		frozen_list=list(reversed(sorted(frozen_list)))
		if debug: print("""
		JCj_region=%s
		Fs_used_list=%s
		frozen_list=%s"""%(JCj_region,Fs_used_list,frozen_list))
		
		quiver={};
		
		overly_expected_arrows={}
		for x,y in expected_arrows:
			if x>=len(Fs_used_list) or y>=len(Fs_used_list):
				continue;
			for xx in Fs_used_list[x]:
				for yy in Fs_used_list[y]:
					if xx==yy: continue
					overly_expected_arrows[(xx-1,yy-1)]=(x,y);
					if xx-1 in frozen_list and yy-1 in frozen_list: continue
					xy=(xx-1,yy-1)
					yx=(yy-1,xx-1)
					if xy not in quiver: quiver[xy]=0; quiver[yx]=0
					quiver[xy]+=1
					quiver[yx]-=1
		self.overly_expected_arrows=overly_expected_arrows;
		#if debug: print("quiver skew symm: %s"%(quiver))
		quiver_pos={}
		for xy in quiver:
			if quiver[xy]>0: quiver_pos[xy]=quiver[xy]
		quiver=quiver_pos
		if debug: print("quiver=%s"%(quiver))
		self.v=v; self.used=used; self.skipped=skipped;
		self.Fs_used_list=Fs_used_list; self.frozen_list=frozen_list; self.JCj_region=JCj_region; self.left_neigh=left_neigh;
		self.expected_arrows=expected_arrows; self.quiver=quiver;

#copied from RQuiver.get_B_matrix()
	def get_RQuiver(self,debug=False,index_add=0,ignore_frozens=False):
		skipped, frozen_list, quiver= self.skipped, self.frozen_list, self.quiver
		remap={};
		remap_i={};
		cnt=0;
		for i in skipped:
			if i not in frozen_list:
				remap[i+index_add]=cnt;
				remap_i[cnt]=i+index_add;
				cnt+=1;
		for i in reversed(frozen_list):
			remap[i+index_add]=cnt;
			remap_i[cnt]=i+index_add;
			cnt+=1;
		if debug: print("skipped=%s, frozen_list=%s, remap=%s"%(skipped,frozen_list,remap));
		n_fro=m=len(frozen_list)
		n_mut=n=len(skipped)-m
		B=Matrix(n+m,n);
		for i,j in quiver:
			ri=remap[i+index_add];
			rj=remap[j+index_add];
			if rj<n: B[ri,rj]+=1;
			if ri<n: B[rj,ri]-=1;
		B.set_immutable();
		return RQuiver(B,labels=remap_i,ignore_frozens=ignore_frozens)


	@staticmethod
	def print_latex(text,file_name="output",tex_dir="tex/",template_path="tex/lec_template.tex",error_from=500,error_to=500,
					return_compil_output=False,clean_aux=True):
		"""print latex using template file
		``error_from`` and ``error_to`` determine neighborhood size of an error message to output when error occurs.
		``template_path`` is path to template file. Must contain "%PASTE%" where we will paste our new text
		output file is found in ``tex_dir/`` + ``file_name``.tex
		""" 
		f=open(template_path,"r");
		template=f.read();
		subbed=template.replace("%PASTE%",text);
		o=open(tex_dir+file_name+".tex","w");
		o.write(subbed);
		o.close()
		command='latexmk -pdf -f -g -bibtex -synctex=1 -halt-on-error -jobname=%s%s %s%s.tex'%(tex_dir,file_name,tex_dir,file_name)
	
		pipe = sp.Popen( command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE )
		res = pipe.communicate()
		compil=res[0].decode("utf-8")
		
		if "error" in compil or "Error" in compil or pipe.returncode!=0:
			print("latex compilation error! return code=",pipe.returncode)
			for m in re.finditer('[Ee]rror', compil):
				print("*************************************\n")
				print(compil[max(0,m.start()-error_from):min(len(compil),m.end()+error_to)])
		else: print("OK! file name="+file_name)
		if clean_aux:
			for ext in ["*.aux","*.log","*.fdb_latexmk","*.fls","*.out","*.blg","*.nav","*.snm","*.toc","*.synctex.gz"]:
				command="rm "+tex_dir+ext
				pipe = sp.Popen( command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE )
				res = pipe.communicate()
# 				if pipe.returncode!=0:
# 					print("ret=%s\nout=%s\n\nerr=%s"%(pipe.returncode,res[0].decode("utf-8"),res[1].decode("utf-8")))
		if return_compil_output: return compil
	
	
	def get_color(self,x,y):
		return str((max(x,y)*71)%100);
	
	def get_arrow_str(self,x,y,index_add=0):
		overly_expected_arrows=self.overly_expected_arrows
		if (x,y) in overly_expected_arrows:
			xx,yy=overly_expected_arrows[(x,y)];
			return str(xx)+"-"+str(x+1+index_add),str(yy)+"-"+str(y+1+index_add);
		else:
			return str(x)+"-"+str(x+1+index_add),str(y)+"-"+str(y+1+index_add);
	
	def get_tpic(self,index_add=0,draw_quiver=True,special=-1,only_mutable_quiver=False,omit_trail=True):
		vred,wred=self.vred,self.wred;
		
# 		n,Fs_used_list, used, quiver, quiver_fro, skipped, left_neigh, expected_arrows, overly_expected_arrows, Fr = self.n, self.Fs_used_list, self.used, self.quiver, self.quiver_fro, self.skipped, self.left_neigh, self.expected_arrows, self.overly_expected_arrows , self.Fr
		
		n,Fs_used_list, used, quiver, skipped, left_neigh, expected_arrows, frozen_list = self.n, self.Fs_used_list, self.used, self.quiver, self.skipped, self.left_neigh, self.expected_arrows, self.frozen_list
		
		wlen=self.wlen
		if omit_trail: wlen=max(skipped)+1
		Fs_str=[" "]*wlen;
	# 	print("wlen=",wlen,") Fs_used_list:",Fs_used_list
		for i in range(wlen):
		# 	print(Fs_used_list[i])
			Fs_str[i]=','.join([str(x+index_add)+"/"+("black!15" if x-1 in frozen_list else "green!50") for x in Fs_used_list[i]])
		xscl=min(0.5,0.5*15/wlen);
		yscl=min(0.8,0.8*6/n)
		tpic="";
		tpic+=("\\begin{tikzpicture}[xscale=%.2f,yscale=%.2f")%(xscl,yscl)+"]\n"
		tpic+=("\\def\\xsh{%.2f}" % (0.5/xscl*0.4));
		tpic+=("\\def\\ysh{%.2f}" % (0.8/yscl*0.08));
		tpic+="\\def\\rbound{"+str(wlen*2+1)+"}\n"
		tpic+="\\wires{"+str(n)+"}\n";
		for i in range(wlen):
			if left_neigh[i]>=0:
				lene=str(left_neigh[i])
			else:
				lene="\\lboundOverTwo";
			#(#4+#2-\xsh*#5+2*\xsh*\i,#3.5-\ysh*#5+2*\ysh*\i)
			# #2=i, #3=wred[i], #4=lene, #5=len(Fs_used_list[i])+1
			if special in Fs_used_list[i]:
				frmt="\\draw[red!40,fill=red!40] (%d,%.2f) circle (0.25cm);\n";
				tpic+=(frmt%(i+left_neigh[i],wred[i]+0.5));
			Fs_str_list=Fs_str[i] if not only_mutable_quiver else ','.join([str(x+index_add)+"/green" 
																			for x in Fs_used_list[i] if x-1 not in frozen_list]);
			Fs_len=len(Fs_used_list[i]) if not only_mutable_quiver else len([x for x in Fs_used_list[i] if x-1 not in frozen_list])
			if i not in used:
				tpic+="\\ledotSEPfilled{"+Fs_str_list+"}("+str(i)+","+str(wred[i])+"){"+lene+"}{"+str(Fs_len)+"}\n";
			else:
				tpic+="\\xxdotSEPfilled{"+Fs_str_list+"}("+str(i)+","+str(wred[i])+"){"+lene+"}{"+str(Fs_len)+"}\n";
		if draw_quiver:
			for x,y in quiver:
				if only_mutable_quiver and (x in frozen_list or y in frozen_list): continue;
				mult=quiver[(x,y)];
				xs,ys=self.get_arrow_str(x,y,index_add=index_add);
				if mult==1:
					if (x in frozen_list or y in frozen_list): dred="drawedgeEXP";
					else: dred="drawedgeTHICK"
					tpic+="\\"+dred+"{"+xs+"}{"+ys+"}{"+self.get_color(x+index_add,y+index_add)+"}\n";
				else:
					tpic+="\\drawedgeMULT{"+xs+"}{"+ys+"}{"+self.get_color(x+index_add,y+index_add)+"}{"+str(mult)+"}\n";
		tpic+="\\end{tikzpicture}\n";
		return tpic+"\n\n\n"

	
	@staticmethod
	def print_quiver(vred, wred, file_name="output",draw_quiver=True, name="", comment="", new_page=False, special=-1,
					 index_add=0, return_tex=False, print_diff=False, prev_quiver=None, print_frozens=False, only_mutable_quiver=False,omit_trail=False):
		"""print a quiver for `vred`, `wred` (need not be reduced!) into file `file_name`.tex.
		parameters:

		* `print_modules`: whether to print the skew shapes.
		* `name`: section title
		* `special`: mark each occurence of this index with a red dot.
		* `index_add`: shift all indices by this value.
		* `return_tex`: if False then we print into `file_name`.tex, otherwise we just return the tikz picture.
		* `only_mutable_quiver`: if True, ignore arrows connecting mutables to frozens
		* `print_diff`: print new/lost arrows/frozens compared to `prev_quiver`."""
		#global FILE_NAME, RETURN_LATEX, v, w, quiver, frozen_list, prev_quiver_local;
		global prev_quiver_local
		tpics="";
		if new_page: tpics+="\\newpage"
		ld=LecDataBraid(vred,wred)
		v, quiver, frozen_list=ld.v, ld.quiver, ld.frozen_list
		if v.length()!=len(vred):
			raise Exception("non-reduced word! len(v)=%d, len(vred)=%d"%(v.length(), len(vred)))
		spec=special;
		if name!=None: tpics+="\\section{%s}\n\n"%(name)
# 		tpics+="\\begin{align*} v&=%s\\\\ w&=%s \\end{align*}"%(vred,wred)
		wred_str="["+ ",".join(["\\textcolor{%s}{%s}"%("black" if j in ld.skipped else "black!30",wred[j]) for j in range(len(wred))]) +"]"
		tpics+="\\begin{align*} v&=\\textcolor{black!30}{%s}\\\\ w&=%s \\end{align*} %s\n\n"%(vred,wred_str,comment)
		arrows="mutable arrows involving %d: "%spec;
		arrs=[]
		tpics+=ld.get_tpic(index_add=index_add,draw_quiver=draw_quiver,special=spec,only_mutable_quiver=only_mutable_quiver, omit_trail=omit_trail)
		for i,j in quiver:
			if i+1==spec and j not in frozen_list:
				arrs.append(" $(%d\\to%d)$ "%(i+1,j+1))
			if j+1==spec and i not in frozen_list:
				arrs.append(" $(%d\\to%d)$ "%(i+1,j+1))
		if spec!=-1:
			tpics+=arrows+",".join(arrs)+"\n"
		if print_frozens: tpics+="frozens (%d): %s"%(len(frozen_list),[i+1+index_add  for i in sorted(frozen_list)])
		if prev_quiver==None: prev=prev_quiver_local;
		else: prev=prev_quiver;
		if prev==None: prev_quiver_local=(frozen_list,quiver);
		elif print_diff:
			fro2,qui2=prev;
			tpics+="\n\n lost frozens: %s\n\n"%([i+1 for i in fro2 if i+1 not in frozen_list])
			tpics+="\n\n new frozens: %s\n\n"%([i+1+index_add for i in frozen_list if (i-1 not in fro2)])
			tpics+="lost arrows:";
			for i,j in qui2:
				if (i+1,j+1) not in quiver: tpics+="$(%d\\to%d)$ "%(i+1,j+1);
			tpics+="\n\nextra arrows:";
			for i,j in quiver:
				if (i-1,j-1) not in qui2: tpics+="$(%d\\to%d)$ "%(i+1,j+1);
		if return_tex: return tpics;
		else: LecDataBraid.print_latex(tpics,file_name=file_name)

	@staticmethod
	def print_multiple(datas,file_name="output",**opts):
		"""`datas`=a list of dictionaries of quivers to print.
		each item of `datas` has to contain a "vred" and a "wred" parameters.
		Additionally, it may contain `file_name`, `name`, `new_page`, `index_add`, `special`, etc
		as in the  `print_quiver` function. Default values can be passed into `opts`."""
		global prev_quiver_local;
		prev_quiver_local=None;
		tpics="";
		for dat in datas:
			dat2=copy(opts);
			dat2.update(dat);
			dat2["return_tex"]=True;
			tpics+=LecDataBraid.print_quiver(**dat2);
		LecDataBraid.print_latex(tpics,file_name=file_name)













