from sage.all import *
import re

class MatrixTools:
	def __init__(self, n=None, auto_simplify=True, expand=False, factor=False, normalize=True):
		"""``n`` is the default size of matrices and permutations.
		``auto_simplify=True`` means we simplify each matrix that we output
		``expand``, ``factor``, ``normalize`` is what we automatically do to the entries of matrices"""
		self.auto_simplify, self.expand, self.factor, self.normalize = auto_simplify, expand, factor, normalize
		if n != None:
			self.init_perms(n);
		else:
			self.n = None;

	def auto_simplify_matrix(self, M):
		"""Simplify matrix ``M`` if ``self.auto_simplify`` is ``True``"""
		if self.auto_simplify:
			return self.simplify_matrix(M);
		return M;

	def dot(self, u, debug=False):
		"""Return a signed permutation matrix of ``u`` whose bottom/left flag minors are TNN.
		This is what is denoted ``\\dot u`` in the paper."""
		ured = self.get_red(u);
		if debug: print("ured=%s" % (ured));
		M = Matrix.identity(SR, self.n);
		for i in range(len(ured)):
			M = M * self.dot_s(ured[i]);
		return M;

	def dot_transpose(self, v):
		"""transposed dot of `v^{-1}`"""
		return self.dot(~v).transpose();

	def FZ_line(self, u):
		"""\\overline{u}, which is same as ``self.dot(u)``"""
		return self.dot(u);

	def FZ_line2(self, u):
		"""double ``\\overline{u}``. Same as ``self.dot_transpose(v)``."""
		return self.dot_transpose(u)

	def init_perms(self, n):
		"""Initialize the permutations and the simple reflections."""
		self.n = n
		self.W = Permutations(n);
		self.s = self.W.simple_reflections();
		self.id = self.W.one()

	def from_red(self, vred):
		"""Convert a reduced word (list) ``vred`` into a permutation."""
		if len(vred) > 0 and max(vred) + 1 > self.n:
			raise Exception("permutation group %s too small for vred=%s" % (self.n, vred));
		v = self.id;
		for i in vred:
			v = v * self.s[i];
		return v;

	def get_red(self, w):
		"""Get the reduced word of a permutation ``w``. This will be reversed from the standard permutation method."""
		return list(reversed(w.reduced_word()));

	def tostrw(self, w):
		"""Pretty print a reduced word of ``w``."""
		if (w.length() == 0):
			return "id";
		answ = "";
		for i in self.get_red(w):
			answ = answ + "s" + str(i) + "*";
		return answ[:-1];

	def tostrwred(self, wred):
		"""Pretty print a reduced word ``wred``."""
		if (len(wred) == 0):
			return "id";
		answ = "";
		for i in wred:
			answ = answ + "s" + str(i) + "*";
		return answ[:-1];

	def get_positive_subexpr(self, v=None, vred=[], w=None, wred=None, check_id_at_the_end=True):
		"""Returns a positive distinguished subexpression of ``v`` inside a braid word ``wred``.
		If ``w`` is specified then we set ``wred`` to be its reduced word.
		If ``vred`` is specificed then we set ``v`` to be the corresponding element of ``W``.
		"""
		if wred == None: wred = self.get_red(w);
		if v == None: v = self.from_red(vred);
		s = self.s;
		indices = [];
		sis = [];
		m = len(wred);
		vv = v;
		if vv.length() == 0:
			return indices, sis;
		for i in range(m - 1, -1, -1):
			si = s[wred[i]];
			if (vv * si).length() < (vv).length():
				indices = [i] + indices;
				sis = [wred[i]] + sis;
				vv = vv * si;
			if vv.length() == 0:
				return indices, sis;
		if check_id_at_the_end:
			raise Exception("no subexpression for " + self.tostrw(v) + " inside " + self.tostrwred(wred) + "!");
		else:
			print("no subexpression for " + self.tostrw(v) + " inside " + self.tostrwred(wred) + "!");
			return indices,sis

	def dot_s(self, i):
		"""returns ``\\dot s_i``
		As a matrix, it is
		0 & -1
		1 & 0
		in rows/columns ``i`` and ``i+1`` (human notation)."""
		M = Matrix.identity(SR, self.n);
		M[i, i] = 0;
		M[i - 1, i - 1] = 0;
		M[i - 1, i] = -1;
		M[i, i - 1] = 1;
		return M;

	def x(self, i, t):
		"""Upper triangular Chevalley generator."""
		M = Matrix.identity(SR, self.n);
		M[i - 1, i] = t;
		return M;

	def y(self, i, t):
		"""Lower triangular Chevalley generator"""
		M = Matrix.identity(SR, self.n);
		M[i, i - 1] = t;
		return M;

	def xy(self, i, j, t):
		"""Identity matrix with off-diagonal ``(i,j)``-entry equal to ``t``"""
		M = Matrix.identity(SR, self.n);
		M[i - 1, j - 1] = t;
		return M;

	def alpha_check(self, i, t):
		"""Diagonal matrix with `(i,i)` and `(i+1,i+1)` diagonal entries equal to `t` and `1/t`"""
		M = Matrix.identity(SR, self.n);
		M[i - 1, i - 1] = t;
		M[i, i] = 1 / t;
		return M;

	def simplify_Laurent_poly(self, P):
		"""Normalize/factor/expand ``P``"""
		answ = SR(P);
		if self.normalize:
			answ = answ.normalize()
		if self.factor:
			answ = answ.factor()
		if self.expand:
			answ = answ.expand()
		else:
			return answ;

	def simplify_matrix(self, M):
		"""Normalize/factor/expand each entry of matrix ``M``"""
		k, n = M.dimensions();
		MM = Matrix(SR, k, n)
		for i in range(k):
			for j in range(n):
				MM[i, j] = self.simplify_Laurent_poly(M[i, j]);
		return MM;


	def matrix_custom_tostr(self, M, return_lines=False):
		"""Some technical method for printing matrices."""
		k, n = M.dimensions();
		maxlen = [0] * n;
		for i in range(k):
			for j in range(n):
				maxlen[j] = max(maxlen[j], len(str(M[i, j])));
		answ = "";
		lines=[]
		for i in range(k):
			line=""
			for j in range(n):
				vert="|" if j<n-1 else ""
				line+=(("%" + str(maxlen[j]) + "s "+vert) % M[i, j])
			answ += line+"\n"
			lines.append(line)
		if return_lines: return lines
		return answ;

	
	def pmat(self, M, *name):
		"""Print a matrix ``M``"""
		nam = "";
		if len(name) > 0:
			nam = str(name[0]) + ":";
		print(nam + "\n" + self.matrix_custom_tostr(self.simplify_matrix(M)), "\n")

	def pmats(self, *mats):
		"""print a bunch of matrices"""
		cnt = 1;
		for M in mats:
			print("%d:\n%s\n\n" % (cnt, self.matrix_custom_tostr(self.auto_simplify_matrix(M))))
			# 		print(cnt,":\n",matrix_custom_tostr(simplify_matrix(M)),"\n\n")
			cnt += 1;


	def pmt(self, name):
		"""Print a matrix where `name` is a string denoting the corresponding sage variable"""
		self.pmat(eval(preparse(name)), name);

	def pmts(self, *names, row=False):
		"""Print a bunch of matrices where `names` is a list of strings denoting the corresponding sage variables
		it's ok for some names to look like "A(x,y),B,C", which is equivalent to "A(x,y)","B","C"
		"""
		names2 = [];
		r = re.compile(
			r'(?:[^,(]|\([^)]*\))+')  # https://stackoverflow.com/questions/1648537/how-to-split-a-string-by-commas-positioned-outside-of-parenthesis/13573834
		for name in names:
			names2 += r.findall(name)
		if not row:
			for name in names2:
				self.pmat(eval(preparse(name)), name);
		else:
			sep=" $ "
			row_lines=[""]
			total_ell=0
			for name in names2:
				M=eval(preparse(name))
				lines=self.matrix_custom_tostr(self.auto_simplify_matrix(M), return_lines=True)
				ell=max(len(lines[0]),len(name))
				pattern="%"+str(ell)+"s"
				row_lines[0]+=pattern%name+sep
				for j in range(len(row_lines),len(lines)+1):
					row_lines.append(" "*total_ell)
				for j in range(len(lines)):
					row_lines[j+1]+=pattern%lines[j]+sep
				total_ell+=ell+len(sep)
			#print("row lines: %s"%row_lines)
			print("\n".join(row_lines))

	def mnr(self, x, rows, cols, sort=True,subtract_one=False):
		"""Compute and normalize a minor of a matrix.
		if ``subtract_one`` then will subtract 1 from all indices"""
		if subtract_one: rows,cols=[i-1 for i in rows],[i-1 for i in cols]
		if sort: rows, cols = sorted(rows), sorted(cols)
		return x.matrix_from_rows_and_columns(rows, cols).det().normalize()


	def w0(self):
		"""Longest element."""
		return Permutation([self.n - i for i in range(self.n)])


	def id_perm(self):
		"""Identity element."""
		return Permutation([i+1 for i in range(self.n)])

	def dw0(self):
		"""``\\dot{w_0}``"""
		return self.dot(self.w0())


	def mat_to_perm(self,M):
		"""``M`` is a permutation matrix. We return the corresponding permutation."""
		n=M.nrows();
		M=self.auto_simplify_matrix(M)
		perm=[]
		for i in range(n):
			nonzero=[]
			for j in range(n):
				if M[i,j]!=0: nonzero.append(j)
			if len(nonzero)!=1: raise Exception("M is not a permutation matrix! i=%d, j=%d, \nM=%s"%(i,j,M))
			perm.append(nonzero[0]+1)
		return Permutation(perm)


	def demazure_product(self,lst):
		"""Here ``lst`` is a braid word given as a list. Return its Demazure product."""
		answ=self.id_perm()
		for i in lst:
			answ_next=answ*self.s[i]
			if answ_next.length()>answ.length():
				answ=answ_next
		return answ

	def z(self,i,t):
		"""Matrix ``z_i(t)``"""
		return self.x(i,t) * self.dot_s(i)

	def zb(self,i,t):
		"""Matrix ``\\bar z_i(t)``"""
		return self.x(i,-t) * ~self.dot_s(i)
