


class RQuiver:
	"""my own class for quivers. Set `seed={}` to turn off seeds."""
	def __init__(self,B,labels={},seed={},MR_mat=None,history=[],ignore_frozens=False,d_sym=None,check_skew_symm=False):
		"""`B` is B-matrix of size (mut+fro times mut). 
		`labels`: names of vertices. 
		`seed`: some functions assigned to vertices.
		`d_sym`: dic of diagonal entries of the skew-symmetrizing matrix. Keys = `range(n)`
		`MR_mat` is the Marsh-Rietsch matrix (specific to Richardson quivers).
		"""
		if ignore_frozens:
			nm,n=B.dimensions()
			B=B.matrix_from_rows(range(n))
		self.B=B
		if d_sym==None:
			self.d_sym={i:1 for i in range(n)}
		else:
			self.d_sym=d_sym
		if B.is_mutable(): self.B.set_immutable()
		self.seed=seed
		self.MR_mat=MR_mat
		nm,n=B.dimensions()
		self.n_mut=n
		self.n_fro=nm-n
		self.n_verts=nm
		if labels=={}:
			self.labels={i:i for i in range(nm)}
		else:
			self.labels=labels
		self.history=history
		if check_skew_symm: self.check_skew_symm()
	#
	# def get_mutable_part(self):
	# 	n=self.n_mut
	# 	B=self.B
	# 	B2=B.matrix_from_rows(range(n))
	#

	def check_skew_symm(self):
		n,d,B=self.n_mut,self.d_sym,self.B
		for i in range(n):
			for j in range(i,n):
				if B[i,j]*d[i] != -B[j,i]*d[j]:
					raise Exception("Matrix\n%s\n is not skew-symmetrized by d=%s!"%(B,d))

	def __str__(self):
		return str(self.history)
	def __repr__(self):
		return self.__str__()
	
	@staticmethod
	def mutate_B_matrix(B,k):
		n,m=B.dimensions()
		B2=Matrix([[B[i,j] for j in range(m)] for i in range(n)]) #somehow copy doesn't work
		for i in range(n):
			for j in range(m):
				if i==k or j==k: B2[i,j]=-B[i,j]
				elif B[i,k]>0 and B[k,j]>0: B2[i,j]=B[i,j]+B[i,k]*B[k,j]
				elif B[i,k]<0 and B[k,j]<0: B2[i,j]=B[i,j]-B[i,k]*B[k,j]
				else: B2[i,j]=B[i,j]
		B2.set_immutable()
		return B2
	
	@staticmethod
	def get_n_edges(B):
		nm,n=B.dimensions()
		summa=0
		for i in range(nm):
			for j in range(n):
				summa+=abs(B[i,j])
		return summa
	
	def get_deg4_verts(self,at_most_4=False,check_2x2=False):
		"""Returns all degree=4 vertices if `at_most_4=False`. 
		If `check_2x2` then check that each deg-4 vertex has 2 in and 2 out"""
		answ=[]
		for i in range(self.n_mut):
			e_in=0
			e_out=0
			for j in range(self.n_verts):
				b=self.B[j,i]
				if b>0: e_out+=b
				else: e_in-=b
			if at_most_4:
				if e_in+e_out>4:continue
			else:
				if (e_in+e_out)!=4: continue
				if (e_in!=2 or e_out!=2) and check_2x2:
					raise Exception("vertex %d is of deg 4 but not alternating!"%i)
			answ.append(i)
		return answ

	def relabel(self,label_to_label_dic):
		dic=label_to_label_dic
		lab=self.labels
		for i in lab:
			if lab[i] in dic: lab[i]=dic[lab[i]]
		self.labels=lab
	
	def from_label(self,label):
		for i in self.labels:
			if self.labels[i]==label: return i
		raise Exception("label %s not found! labels=%s"%(label,self.labels))
	
	def Bij_from_label(self,label_i, label_j):
		i,j=self.from_label(label_i),self.from_label(label_j)
		d=self.d_sym
		if j<self.n_mut: return self.B[i,j]
		else: return -d[j]/d[i] * self.B[j,i]
		#print("RQUIVER: i=%d, j=%d, Bdim=%s"%(i,j,self.B.dimensions()))
		# return self.B[i,j]

	def mutate_several(self,indices=None,labels=None):
		if indices!=None: raise NotImplemented()
		if labels!=None:
			prev=self
			for k_label in labels:
				new_label=None
				if isinstance(labels,dict): new_label=labels[k_label]
				prev=prev.mutate(k_label=k_label,new_label=new_label)
				#print("mutated at %s\nnew labels=%s"%(k_label,prev.labels))
			return prev
				
	def mutate(self,k=None,k_label=None,new_label=None):
		"""Usual CA mutation. If `seed={}` then only mutate the quiver."""
		if k==None:
			k=self.from_label(k_label)
			if k>=self.n_mut: raise Exception("k_label=%s; k=%s; frozen! all labels=%s"%(k_label,k,self.labels))
			#print("k_label=%s; k=%s"%(k_label,k))
		B2=self.mutate_B_matrix(self.B,k)
		seed2=copy(self.seed)
		if self.seed!={}:
			prod_in=1
			prod_out=1
			for i in range(self.n_verts):
				b=self.B[i,k]
				if b>0: prod_in*=self.seed[i]^b
				elif b<0: prod_out*=self.seed[i]^-b
			seed2[k]=((prod_in+prod_out)/self.seed[k]).normalize().factor()
		#print("seed2=(%s+%s)/(%s)=%s"%(prod_in,prod_out,self.seed[k],seed2[k]))
		labels2=copy(self.labels)
		if new_label==None and self.labels!={}:
			new_label="%s'"%(self.labels[k])
			labels2[k]=new_label
		return RQuiver(B=B2,labels=labels2,seed=seed2,MR_mat=self.MR_mat,history=self.history+[k],d_sym=self.d_sym)
	
# 	def show(self,*opts):
# 		ClusterQuiver(self.B).show(*opts)
	
	def __eq__(self,other):
		if other==None: return False
		return (self.B==other.B and self.seed==other.seed)
	
	def __hash__(self):
		return hash(self.B)+hash(tuple(list(self.seed)))
	
	def __get_seed_inv(self):
		return {self.seed[i]:i for i in self.seed}
	
	def is_isomorphic(self,other,debug=False,return_permutation=False,ignore_frozens=False):
		"""if `return_permutation` then we return a dic when the isomorphism is found"""
		if ignore_frozens:
			if self.n_mut != other.n_mut: return False
			n=self.n_mut
			B1=self.B.matrix_from_rows(range(n))
			B2=other.B.matrix_from_rows(range(n))
			q1=RQuiver(B1)
			q2=RQuiver(B2)
			return q1.is_isomorphic(q2,debug=debug, return_permutation=return_permutation,ignore_frozens=False)
		if self.seed=={}:
			if self.n_mut!=other.n_mut or self.n_fro!=other.n_fro: return False
			row_to_indices={}
			Bs=[self.B,other.B]
			for i in range(self.n_verts):
				for j in range(2):
					row=tuple(sorted(Bs[j][i]))
					if row not in row_to_indices: row_to_indices[row]=([],[])
					row_to_indices[row][j].append(i)
			prev=[{}]
			for row in row_to_indices:
				if len(row_to_indices[row][0])!=len(row_to_indices[row][1]): return False
				r=len(row_to_indices[row][0])
				nxt=[]
				for u in Permutations(range(r)):
					for v in prev:
						uv=copy(v)
						is_good=True
						for i in range(r):
							x1=row_to_indices[row][0][i]
							x2=row_to_indices[row][1][u[i]]
							if (x1<self.n_mut)!=(x2<self.n_mut): is_good=False; break
							uv[x1]=x2
						if is_good: nxt.append(uv)
				prev=nxt
			row_to_indices
			prev
			for p in prev:
				is_good=True
				for i in range(self.n_verts):
					for j in range(self.n_mut):
						if self.B[i,j]!=other.B[p[i],p[j]]:
							is_good=False
							break
					if not is_good: break
				if is_good:
					if return_permutation: return p
					else: return True
			return False
		si=self.__get_seed_inv()
		Qsi=Q.__get_seed_inv()
		if sorted(list(si),key=str)!=sorted(list(Qsi),key=str):
			if debug:
				print("diff seeds! %s   !=   %s"%(sorted(list(si)),sorted(list(Qsi))) )
			return False
		perm={si[s]:Qsi[s] for s in si}
		if debug: print("perm = %s"%perm)
		for i in range(self.n_verts):
			for j in range(self.n_mut):
				if self.B[i,j]!=Q.B[perm[i],perm[j]]:
					if debug: print("i=%d, j=%d; Bij=%d  perm[i]=%d; perm[j]=%d, Bpipj=%d"%(i,j,self.B[i,j],perm[i],perm[j],Q.B[perm[i],perm[j]]))
					return False
		if return_permutation: return perm
		return True
	
	def get_all_mutations(self, exit_function=None, max_it=1000, debug=False, max_depth=100, find_all=False, only_deg_4=False, deg_at_most_4=False, up_to_iso=True):
		"""Apply some mutations. Return `Q` if `exit_function(Q)=True`. 
		Otherwise keeps looking and returns the list of all quivers.
		By default (unless `only_deg_4` or `deg_at_most_4`), mutate all possible vertices.
		`max_it`=maximal number of iterations of the bfs.
		`up_to_iso`=up to isomorphism. Should be always `True`."""
		Q=self
		if exit_function!=None and exit_function(Q):
			if find_all: return [Q]
			else: return Q
		queue=[Q]
		used={Q:True}
		depth={Q:0}
		cnt=0
		found_list=[]
		while cnt<len(queue):
			if debug: print(cnt,end=",")
			Q=queue[cnt]
			cnt+=1
			if depth[Q]>max_depth: continue
			deg4=range(Q.n_mut)
			if only_deg_4: deg4=Q.get_deg4_verts(at_most_4=deg_at_most_4)
			for k in deg4:
				Q2=Q.mutate(k)
				if up_to_iso:
					found=False
					for Qu in used:
						if Q2.is_isomorphic(Qu):
							found=True
							break
				else: found=(Q2 in used)
				if not found:
					queue.append(Q2)
					used[Q2]=True
					depth[Q2]=depth[Q]+1
					if exit_function!=None and exit_function(Q2):
						if not find_all: return Q2
						found_list.append(Q2)
						max_depth=min(max_depth,depth[Q2])
			if cnt>max_it:
				#print("too many iterations! breaking...")
				break
		if exit_function!=None and find_all:
			if debug: print("\nOK! found %d quivers!" % (len(found_list)))
			return found_list
		if debug: print("\nOK! found %d quivers!"%(len(queue)))
		if exit_function!=None: return None; #desired quiver not found
		return queue
	
	def get_mutable_abs_degrees(self):
		"""return a dic, for each mutable vertex the total number of arrows incident to it"""
		degrees={}
		for i in range(self.n_mut):
			degrees[i]=0
			for j in range(self.n_mut):
				degrees[i]+=abs(self.B[i,j])
		return degrees
	
	def has_sep_edge(self,return_one=False, return_all=False, debug=False, leaf_only=False):
		"""Return whether `self` has a separating edge in the sense of Muller.
		if ``return_one`` then return one edge if found.
		if ``return_all`` then return all edges.
		if ``leaf_only`` then must be separating leaf"""
		edges=[]
		for i in range(self.n_mut):
			for j in range(self.n_mut):
				if self.B[i,j]>0: edges.append((i,j))
		D=DiGraph(edges)
		Dscc=D.strongly_connected_components_digraph()
		scc=D.strongly_connected_components()
		if debug: print("strongly conn cpts=",scc)
		has_left_inf={}
		has_right_inf={}
		for cpt in Dscc.vertices():
			if len(cpt)>1:
				has_left_inf[cpt]=True
				has_right_inf[cpt]=True
		relaxed=True
		while relaxed:
			relaxed=False
			for cpt1,cpt2,nun in Dscc.edges():
				if cpt1 in has_left_inf and cpt2 not in has_left_inf:
					has_left_inf[cpt2]=True
					relaxed=True
				if cpt1 not in has_right_inf and cpt2 in has_right_inf:
					has_right_inf[cpt1]=True
					relaxed=True

		has_left_inf_v=[v for cpt in has_left_inf for v in cpt]
		has_right_inf_v=[v for cpt in has_right_inf for v in cpt]
		if debug: print("LEFT :",has_left_inf_v)
		if debug: print("RIGHT:",has_right_inf_v)
		all_edges=[]
		degrees=self.get_mutable_abs_degrees()
		for u,v,nun in D.edges():
			if leaf_only and not (degrees[u]==1 or degrees[v]==1): continue
			if (u not in has_left_inf_v) or (v not in has_right_inf_v):
				if return_all: all_edges.append((u,v))
				elif return_one: return (u,v)
				else: return True
		if len(all_edges)==0: return False
		return all_edges
	
	def remove_vertex(self,v):
		relabel=[i for i in range(v)]+[None]+[i-1 for i in range(v+1,self.n_verts)]
		if v>=self.n_mut:
			B2=Matrix(self.n_verts-1,self.n_mut)
		else:
			B2=Matrix(self.n_verts-1,self.n_mut-1)
		for i in range(self.n_verts):
			if i==v: continue
			for j in range(self.n_mut):
				if j==v: continue
				B2[relabel[i],relabel[j]]=self.B[i,j]
		labels2={relabel[i]:self.labels[i] for i in self.labels if i!=v}
		seed2={relabel[i]:self.seed[i] for i in self.seed if i!=v}
		return RQuiver(B=B2, labels=labels2,seed=seed2,history=self.history+[-v])
	
	
	
	def get_point_count(self,max_it=1000,debug=False, depth=0):
		"""if self has exactly the right number of frozens to be full rank then the answer is correct.
		Otherwise it is multiplied by (q-1)^{n_fro-correct_n_fro}. The exponent may be negative
		Base case: A1+1fro gives (q^2-q+1). 
		Thus A1+0fro gives (q^2-q+1)/(q-1), and A1+2fro gives (q^2-q+1)*(q-1)
		`depth` is how deep we are into the recurrence"""
		has_edges=False
		for i in range(self.n_mut):
			for j in range(self.n_mut):
				if self.B[i,j]!=0: has_edges=True; break
			if has_edges: break
		q=var("q")
		if not has_edges:
			return (q**2-q+1)**self.n_mut*(q-1)**(self.n_fro-self.n_mut)
		Qs=self.get_all_mutations(exit_function=lambda Q: Q.has_sep_edge(),max_it=max_it)
		if Qs==None:
			raise Exception("couldn't find a separating edge!")
		u,v=Qs.has_sep_edge(return_one=True)
		if debug:
			#Qs.show(figsize=3)
			print("sep edge:",Qs.labels[u],"->",Qs.labels[v])
		um,vm=min(u,v),max(u,v)
		Qv=Qs.remove_vertex(vm)
		Qu=Qs.remove_vertex(um)
		Quv=Qv.remove_vertex(um)
		Qupc=Qu.get_point_count(max_it=max_it,debug=debug, depth=depth+1)
		Qvpc=Qv.get_point_count(max_it=max_it,debug=debug, depth=depth+1)
		Quvpc=Quv.get_point_count(max_it=max_it,debug=debug, depth=depth+1)
		answ=(q-1)*(Qvpc+Qupc-(q-1)*Quvpc)
		if depth==0:
			answ=answ.normalize()
		if debug:
			print("Qvpc=%s\nQuvpc=%s\nansw=%s"%(Qvpc,Quvpc,answ))
		return answ
	
	
	def get_homfly(self,max_it=100,debug=False, depth=0, check_well_defined=True, return_number_of_ways=True):
		"""Compute conjectural homfly polynomial of a quiver.
		We're only using separating leaves!
		The specialization is `a=q^{-1/2}`, `z=q^{1/2}-q^{-1/2}`.
		if self has exactly the right number of frozens to be full rank then the answer is correct.
		Otherwise it is multiplied by (q-1)^{n_fro-correct_n_fro}. The exponent may be negative
		Base case: A1+1fro gives (q^2-q+1). 
		Thus A1+0fro gives (q^2-q+1)/(q-1), and A1+2fro gives (q^2-q+1)*(q-1)
		`depth` is how deep we are into the recurrence"""
		global Q_problem_1,Q_problem_2
		has_edges=False
		for i in range(self.n_mut):
			for j in range(self.n_mut):
				if self.B[i,j]!=0: has_edges=True; break
			if has_edges: break
		a,z=SR("a"), SR("z")
		homA1fro1=(z+1/z)/a-1/a^3/z
# 		extra_fro=z/a
# 		extra_fro=(a-1/a)/z
		extra_fro=1
		if not has_edges:
# 			return (q^2-q+1)^self.n_mut*(q-1)^(self.n_fro-self.n_mut)
			answ=homA1fro1^self.n_mut*extra_fro^(self.n_fro-self.n_mut)
			if return_number_of_ways: return (answ,1)
			return answ
		uvs=[]
		if not check_well_defined:
			Qs=self.get_all_mutations(exit_function=lambda Q: Q.has_sep_edge(leaf_only=True),max_it=max_it)
			if Qs==None:
				raise Exception("couldn't find a separating edge!")
			u,v=Qs.has_sep_edge(return_one=True,leaf_only=True)
			degrees=Qs.get_mutable_abs_degrees()
			if degrees[u]!=1:
				t=u; u=v; v=t
			assert(degrees[u]==1)
			if debug:
				#Qs.show(figsize=3)
				print("sep edge:",Qs.labels[u],"->",Qs.labels[v])
			uvs=[(u,v,Qs)]
		else:
			Qs_all=self.get_all_mutations(max_it=max_it)
			uvs_dic={}
			for Qs in Qs_all:
				degrees=Qs.get_mutable_abs_degrees()
				edges=Qs.has_sep_edge(return_all=True,leaf_only=True)
				if not edges: continue
				for u,v in edges:
					if degrees[u]!=1: t=u; u=v; v=t
					assert(degrees[u]==1)
					uvs_dic[u,v]=Qs
			uvs=[(u,v,uvs_dic[u,v]) for u,v in uvs_dic]
			if len(uvs)==0:
				raise Exception("couldn't find a separating leaf!")
# 			if debug:
# 				print("sep edges:",Qs.labels[u],"->",Qs.labels[v])
		prev_answ=None
		prev_uvQs=None
		answ_cnt=0
		for u,v,Qs in uvs:
			Qv=Qs.remove_vertex(v)
			Qu=Qs.remove_vertex(u)
			Quv=Qv.remove_vertex(u if u<v else u-1)
			if debug: print("depth=%d; removing u=%d..."%(depth,u))
			Qupc,nw1=Qu.get_homfly(check_well_defined=check_well_defined, max_it=max_it, debug=debug, depth=depth+1, return_number_of_ways=True)
			if debug: print("depth=%d; removing u=%d and v=%d..."%(depth,u,v))
			Quvpc,nw3=Quv.get_homfly(check_well_defined=check_well_defined, max_it=max_it, debug=debug, depth=depth+1, return_number_of_ways=True)
			answ=z/a*Qupc+1/a^2*Quvpc
			answ_cnt+=nw1*nw3
			if depth==0:
				answ=answ.normalize()
			if debug:
				print("depth=%d;  Qupc=%s\nQuvpc=%s\nansw=%s"%(depth,Qupc.normalize(),Quvpc.normalize(),answ.normalize()))
			if prev_answ==None:
				prev_answ=answ
				prev_uvQs=(u,v,Qs)
			if debug: print("prev answ=%s, answ=%s"%(prev_answ,answ))
			elif check_well_defined and not bool(prev_answ==answ):
				u1,v1,Qs1=prev_uvQs
				print("u1=%s, v1=%s, Q1=%s\nu2=%s, v2=%s, Q2=%s\nansw1=%s\nansw2=%s"%(u1,v1,Qs1,u,v,Qs,prev_answ,answ))
				Qs1.show()
				Qs.show()
				Q_problem_1,Q_problem_2=Qs1,Qs
				raise Exception("not well defined!")
			if debug: print("check ok!")
		if return_number_of_ways: return (answ,answ_cnt)
		return answ
	
	def get_mutable_deg(self,i):
		answ=0
		if i>=self.n_mut: return 0
		for j in range(self.n_mut):
			answ+=abs(self.B[i,j])
		return answ
	
	
	def get_all_deg4_mutations_old(self,max_it=1000,debug=True,mutate_in_all_directions=False, at_most_4=False, up_to_iso=True, termination_fcn=lambda Q:False, termination_all_min_length=True):
		"""old function, need to merge to get_all_mutations
		if `mutate_in_all_directions` is `True` then `at_most_4` doesn't matter"""
		global terminated_Q,terminated_Qs
		terminated_Q=None
		terminated_Qs=[]
		Q=self
		queue=[Q]
		used={Q:True}
		cnt=0
		while cnt<len(queue):
			if debug and cnt%10==0: print(cnt,end=",")
			Q=queue[cnt]
			cnt+=1
			if terminated_Q!=None and len(terminated_Q.history)<len(Q.history):
				return queue
			if terminated_Q!=None and termination_fcn(Q):
				terminated_Qs.append(Q)
			deg4=Q.get_deg4_verts(at_most_4=at_most_4)
			if mutate_in_all_directions:
				deg4=range(Q.n_mut)
			for k in deg4:
				Q2=Q.mutate(k)
				if up_to_iso:
					found=False
					for Qu in used:
						if Q2.is_isomorphic(Qu):
							found=True
							break
				else: found=(Q2 in used)
				if not found:
					queue.append(Q2)
					used[Q2]=True
					#print(Q2,"->",termination_fcn(Q2))
					if termination_fcn(Q2):
						if terminated_Q==None:
							print("found termination Q=",Q2)
							terminated_Qs.append(Q2)
							terminated_Q=Q2
						if not termination_all_min_length:
							return queue
			if cnt>max_it:
				print("too many iterations! breaking...")
				break
		if debug: print("\nOK! found %d quivers!"%(len(queue)))
		return queue
	
	
	def show(self,**opts):
		"""show my digraph"""
		opts["partition"]=[[self.labels[i] for i in range(self.n_mut)],[self.labels[i] for i in range(self.n_mut,self.n_verts)]]
		if "use_labels" in opts:
			D=self.get_digraph(use_labels=opts["use_labels"])
			if not opts.pop("use_labels"):
				opts["partition"]=[[i for i in range(self.n_mut)],[i for i in range(self.n_mut,self.n_verts)]]
		else:
			D=self.get_digraph()
		D.show(**opts)

	def get_edges(self,use_labels=True):
		edges=[]
		labels=self.labels if use_labels else {i:i for i in range(self.n_verts)}
		for i in range(self.n_verts):
			for j in range(self.n_mut):
				if self.B[i,j]>0: edges+=[(labels[i],labels[j])]*self.B[i,j]
				if i>=self.n_mut and self.B[i,j]<0: edges+=[(labels[j],labels[i])]*(-self.B[i,j])
		return edges
	
	def get_digraph(self,use_labels=True):
		edges=self.get_edges(use_labels=use_labels)
		verts=[]
		labels=self.labels if use_labels else {i:i for i in range(self.n_verts)}
		for i in range(self.n_verts):
			verts.append(labels[i])
		D=DiGraph([verts,edges],multiedges=True)
		return D
	
	@staticmethod
	def get_B_matrix(lec_data,return_remap=False,debug=False):
		""" ``lec_data`` is an element of ``LecData`` class
		if ``return_remap`` then we return not only B-matrix but also ``remap`` and ``remap_i``"""
		ld=lec_data
		skipped, frozen_list, quiver= ld.skipped, ld.frozen_list, ld.quiver
		remap={}
		remap_i={}
		cnt=0
		for i in skipped:
			if i not in frozen_list:
				remap[i]=cnt
				remap_i[cnt]=i
				cnt+=1
		for i in reversed(frozen_list):
			remap[i]=cnt
			remap_i[cnt]=i
			cnt+=1
		if debug: print("skipped=%s, frozen_list=%s, remap=%s"%(skipped,frozen_list,remap))
		n_fro=m=len(frozen_list)
		n_mut=n=len(skipped)-m
		B=Matrix(n+m,n)
		for i,j in quiver:
			ri=remap[i]
			rj=remap[j]
			if rj<n: B[ri,rj]+=1
			if ri<n: B[rj,ri]-=1
		B.set_immutable()
		if return_remap: return (B,remap,remap_i)
		else: return B





def get_generic_M(n):
	M=Matrix(SR,n)
	for i in range(n):
		M[i,i]=1
		for j in range(i+1,n):
			if n<10: M[i,j]=var("t"+str(i)+str(j))
			else: M[i,j]=var("t"+str(i)+"_"+str(j))
	return M

def get_GBm_M(v,wred):
	M=~dot_t(v) * get_Deodhar_GBmred(v,wred)
	U,D,L=get_UDL(M)
	return U


def get_RQuiver(debug=False,use_matrix="twist"):
	"""`use_matrix`="twist", "generic", or "GBm" """
	global v,wred,n,n_mut,n_fro,remap_i,Fr
	if use_matrix=="generic":
		Mtw=get_generic_M(n)
	elif use_matrix=="twist":
		M=get_Deodhar_GBmred(v,wred)
		Mtw=simplify_matrix(~dot_t(v)*LEC_twist(M,v,w))
	elif use_matrix=="GBm":
# 		Mtw=get_Deodhar_GBmred(v,wred) * ~dot_t(v)
# 		Mtw=~dot_t(v) * get_Deodhar_GBmred(v,wred);#this is wrong
		Mtw=get_GBm_M(v,wred)
	else: raise Exception("wrong matrix option: %s"%use_matrix)
	labels={}
	seed={}
	if debug: print("MUTABLE:")
	for i in range(n_mut+n_fro):
		if i==n_mut:
			if debug: print("FROZEN:")
		ri=remap_i[i]
		Ai,Bi=Fr[ri]
		seed[i]=Mtw.matrix_from_rows_and_columns(Ai,Bi).det().expand().factor()
		labels[i]=(Ai,Bi)
		if debug: print("i=%d, r=%d; label=%30s, seed=%s"%(i,ri,labels[i],seed[i]))
	return RQuiver(B,labels,seed,MR_mat=Mtw)

def compute_all_minors(Q,debug=False):
	global mnr_dic,mnr_dic_i
	M=Q.MR_mat
	n=M.nrows()
	mnr_dic={}
	mnr_dic_i={}

	for size in range(1,n):
		if debug: print("size=%d"%size)
		for SS in Subsets(range(n),size):
			S=sorted(list(SS))
			for TT in Subsets(range(n),size):
				T=sorted(list(TT))
				is_good=True
				for i in range(size):
					if S[i]>=T[i]: is_good=False; break
				if not is_good: continue
				mnr=M.matrix_from_rows_and_columns(S,T).det().expand().factor()
				mnr_dic[(tuple(S),tuple(T))]=mnr
				mnr_dic_i[mnr]=(tuple(S),tuple(T))
	if debug: print("OK done!")
	return 

def check_Pluckerness(Qs,debug=False):
	global mnr_dic_i
	cnt=0
	for Q in Qs:
		if Q.is_Plucker(mnr_dic_i,debug=debug):
			if debug: print(Q,"is Plucker")
		else:
			print(Q,"NON PLUCKER!!!")
			return False
		cnt+=1
	return True

def get_red_words_up_to_comm_moves(wred):
	n=max(wred)+1
	w=Permutations(n).from_reduced_word(wred)
	G=w.reduced_word_graph()
	v0=None
	for v in G.vertices():
		if v==tuple(wred):
			v0=v
			break
	if v0==None: raise Exception("somehow red word graph doesn't contain wred o_O")
	comm_moves=[e for e in G.edges() if e[2]==2]
	G2=copy(G)
	G2.contract_edges(comm_moves)
	return [list(v) for v in G2.vertices()]


def guess_minor(mnr,M):
	"""given a polynomial `mnr`, find rows and cols of a generic upper-triang matrix `M` with this minor"""
	ts=mnr.variables()
	rows=[]
	cols=[]
	for tij in ts:
		stij=str(tij)
		if "_" not in stij:
			i,j=int(str(tij)[1]),int(str(tij)[2])
		else:
			ii,jj=tuple(stij[1:].split("_"))
			i,j=int(ii),int(jj)
		rows.append(i)
		cols.append(j)
	rows=sorted(list(set(rows)))
	cols=sorted(list(set(cols)))
	if len(rows)!=len(cols):
		return None
	mnr2=M.matrix_from_rows_and_columns(rows,cols).det().normalize().factor()
	if mnr2!=mnr:
		return None
	else:
		return (rows,cols)


def get_RQuiver_no_seed(debug=False,use_wred_labels=True):
	global v,wred,n_mut,n_fro,remap_i,Fr
	labels={}
	seed={}
	if debug: print("MUTABLE:")
	for i in range(n_mut+n_fro):
		if i==n_mut:
			if debug: print("FROZEN:")
		ri=remap_i[i]
		Ai,Bi=Fr[ri]
		if use_wred_labels: labels[i]=ri+1
		else:labels[i]=i
		if debug: print("i=%d, r=%d; label=%30s"%(i,ri,labels[i]))
	return RQuiver(B=B,labels=labels)









# 	def get_homfly_arbitrary_sep_edge(self,max_it=100,debug=False, depth=0, check_well_defined=True, return_number_of_ways=True):
# 		"""Compute conjectural homfly polynomial of a quiver. Same recurrence as ``get_point_count``.
# 		The specialization is `a=q^{-1/2}`, `z=q^{1/2}-q^{-1/2}`.
# 		if self has exactly the right number of frozens to be full rank then the answer is correct.
# 		Otherwise it is multiplied by (q-1)^{n_fro-correct_n_fro}. The exponent may be negative
# 		Base case: A1+1fro gives (q^2-q+1). 
# 		Thus A1+0fro gives (q^2-q+1)/(q-1), and A1+2fro gives (q^2-q+1)*(q-1)
# 		`depth` is how deep we are into the recurrence"""
# 		global Q_problem_1,Q_problem_2
# 		raise Exception("I don't know the formula for arbitrary sep edge!")
# 		has_edges=False
# 		for i in range(self.n_mut):
# 			for j in range(self.n_mut):
# 				if self.B[i,j]!=0: has_edges=True; break
# 			if has_edges: break
# 		a,z=SR("a"), SR("z")
# 		homA1fro1=(z+1/z)/a-1/a^3/z
# 		extra_fro=z/a
# 		if not has_edges:
# # 			return (q^2-q+1)^self.n_mut*(q-1)^(self.n_fro-self.n_mut)
# 			answ=homA1fro1^self.n_mut*extra_fro^(self.n_fro-self.n_mut)
# 			if return_number_of_ways: return (answ,1)
# 			return answ
# 		uvs=[]
# 		if not check_well_defined:
# 			Qs=self.get_all_mutations(exit_function=lambda Q: Q.has_sep_edge(),max_it=max_it)
# 			if Qs==None:
# 				raise Exception("couldn't find a separating edge!")
# 			u,v=Qs.has_sep_edge(return_one=True)
# 			if debug:
# 				#Qs.show(figsize=3)
# 				print("sep edge:",Qs.labels[u],"->",Qs.labels[v])
# 			uvs=[(u,v,Qs)]
# 		else:
# 			Qs_all=self.get_all_mutations(max_it=max_it)
# 			uvs_dic={}
# 			for Qs in Qs_all:
# 				edges=Qs.has_sep_edge(return_all=True)
# 				if not edges: continue
# 				for uv in edges:
# 					uvs_dic[uv]=Qs
# 			uvs=[(u,v,uvs_dic[u,v]) for u,v in uvs_dic]
# 			if len(uvs)==0:
# 				raise Exception("couldn't find a separating edge!")
# # 			if debug:
# # 				print("sep edges:",Qs.labels[u],"->",Qs.labels[v])
# 		prev_answ=None
# 		prev_uvQs=None
# 		answ_cnt=0
# 		for u,v,Qs in uvs:
# 			um,vm=min(u,v),max(u,v)
# 			Qv=Qs.remove_vertex(vm)
# 			Qu=Qs.remove_vertex(um)
# 			Quv=Qv.remove_vertex(um)
# 			Qupc,nw1=Qu.get_homfly(max_it=max_it,debug=debug, depth=depth+1,return_number_of_ways=True)
# 			Qvpc,nw2=Qv.get_homfly(max_it=max_it,debug=debug, depth=depth+1,return_number_of_ways=True)
# 			Quvpc,nw3=Quv.get_homfly(max_it=max_it,debug=debug, depth=depth+1,return_number_of_ways=True)
# 			answ=(q-1)*(Qvpc+Qupc-(q-1)*Quvpc)
# 			answ_cnt+=nw1*nw2*nw3
# 			if depth==0:
# 				answ=answ.normalize()
# 			if debug:
# 				print("Qvpc=%s\nQuvpc=%s\nansw=%s"%(Qvpc,Quvpc,answ))
# 			if prev_answ==None:
# 				prev_answ=answ
# 				prev_uvQs=(u,v,Qs)
# 			elif check_well_defined and not bool(prev_answ==answ):
# 				u1,v1,Qs1=prev_uvQs
# 				print("u1=%s, v1=%s, Q1=%s\nu2=%s, v2=%s, Q2=%s\nansw1=%s\nansw2=%s"%(u1,v1,Qs1,u,v,Qs,prev_answ,answ))
# 				Qs1.show()
# 				Qs.show()
# 				Q_problem_1,Q_problem_2=Qs1,Qs
# 				raise Exception("not well defined!")
# 		if return_number_of_ways: return (answ,answ_cnt)
# 		return answ
