from sage.all import *
#load("LecDoubleW.py")
#from LecDoubleW import *
import re

class WeylTools:
	def __init__(self, letter=None, rank=None, typ=None,use_dual=False):
		"""Class wrapper for Weyl groups. One can specify a Weyl group, say, D5
		by either setting ``letter="D"`` and ``rank=5`` or ``typ=["D",5]``.
		"""
		if typ==None:
			typ=[letter,rank]
		else:
			letter,rank=tuple(typ)
		self.letter=letter
		self.rank=rank
		self.typ=typ
		self.n=rank
		R=RootSystem(typ)
		if use_dual: R=R.dual
		L=R.ambient_space()
		W=L.weyl_group()
		s=W.simple_reflections()
		alpha=L.simple_roots()
		id=W.unit()
		I=W.index_set()
		w0=W.long_element()
		omega=L.fundamental_weights()
		comega={j:omega[j]/omega[j].scalar(alpha[j]) for j in I}
		self.R,self.L,self.W,self.s,self.alpha,self.id,self.omega,self.comega,self.I,self.w0=\
			R,L,W,s,alpha,id,omega,comega,I,w0
		self.w0_inv=self.w0_involution_dic()
		
	def get_cartan_dic(self):
		"""Return Cartan matrix as a dictionary.
		the ``i,j``-th entry is ``<alpha_i,alpha_j^\\vee>``
		"""
		answ={}
		alpha=self.alpha
		ach=self.L.simple_coroots()
		I=self.I
		for i in I:
			for j in I:
				answ[i,j]=alpha[j].scalar(ach[i])
		return answ
		
	def from_red(self, vred):
		"""Given a reduced word encoded as a list ``vred``, return the corresponding element of ``W``."""
		return self.W.from_reduced_word(vred)

	def get_red(self, w):
		"""Return the reduced word (as a list) of an element ``w`` of ``W``."""
		return w.reduced_word()

	def tostrw(self, w):
		"""Pretty print the reduced word of ``w``"""
		if (w.length() == 0):
			return "id";
		answ = "";
		for i in self.get_red(w):
			answ = answ + "s" + str(i) + "*";
		return answ[:-1];

	def tostrwred(self, wred):
		"""Pretty print a given reduced word ``wred``"""
		if (len(wred) == 0):
			return "id";
		answ = "";
		for i in wred:
			answ = answ + "s" + str(i) + "*";
		return answ[:-1];

	def get_positive_subexpr(self, v=None, vred=[], w=None, wred=None):
		"""Returns a positive distinguished subexpression of ``v`` inside a braid word ``wred``.
		If ``w`` is specified then we set ``wred`` to be its reduced word.
		If ``vred`` is specificed then we set ``v`` to be the corresponding element of ``W``.
		"""
		if wred == None: wred = self.get_red(w);
		if v == None: v = self.from_red(vred);
		s = self.s;
		indices = [];
		sis = [];
		m = len(wred);
		vv = v;
		if vv.length() == 0:
			return indices, sis;
		for i in range(m - 1, -1, -1):
			si = s[wred[i]];
			if (vv * si).length() < (vv).length():
				indices = [i] + indices;
				sis = [wred[i]] + sis;
				vv = vv * si;
			if vv.length() == 0:
				return indices, sis;
		raise Exception("no subexpression for " + self.tostrw(v) + " inside " + self.tostrwred(wred) + "!");

	def w0(self):
		"""Longest element of ``W``."""
		return self.w0

	def unit(self):
		"""Identity of ``W``."""
		return self.id


	def w0_involution_dic(self):
		"""Return the dictionary ``{i:i^*}`` induced by the conjugation by ``w_0``."""
		I,w0,alpha,typ=self.I,self.w0,self.alpha,self.typ
		answ={}
		for i in I:
			if i in answ: continue
			w0a=-w0.action(alpha[i])
			found=False
			for j in I:
				if w0a==alpha[j]:
					found=True
					answ[i]=j
					answ[j]=i
					break
			if not found:
				raise Exception("%s i=%s, w0(alpha_i)=%s -- not found!"%(typ,i,w0a))
		return answ

					
	def demazure_product(self,lst):
		"""Here ``lst`` is a braid word given as a list. Return its Demazure product."""
		#print("computing dem product of %s"%(lst))
		return self.W.demazure_product(lst)

	def demazure_subtract_right(self,x,y,debug=False):
		""" Act on ``x`` in ``W`` on the right by letters of ``y``
		(either ``y`` in ``W`` or y is a reduced word given as a list),
		making it as small as possible."""
		W=self.W
		if y in W:
			y=self.get_red(y)
		prev=x
		if debug: print("       %s"%(self.get_red(prev)))
		for i in y:
			nxt=prev*self.s[i]
			if prev.length()>nxt.length():
				prev=nxt
			if debug: print("i=%d -> %s"%(i,self.get_red(prev)))
		return prev
	
	def demazure_subtract_left(self,y,x):
		""" Act on ``x`` in W on the left by letters of ``y``
		(either ``y`` in ``W`` or ``y`` is a reduced word given as a list),
		making it as small as possible"""
		W=self.W
		if y in W:
			y=self.get_red(y)
		prev=x
		for i in reversed(y):
			nxt=self.s[i]*prev
			if prev.length()>nxt.length():
				prev=nxt
		return prev
	
	

	def simple_roots_expansion_dict(self, v):
		"""write vector v in terms of simple roots. Return coefficients as dictionary."""
		return {j:self.comega[j].scalar(v) for j in self.I}

	def simple_roots_expansion_list(self, v):
		"""write vector v in terms of simple roots. Return coefficients as a list."""
		a=self.simple_roots_expansion_dict(v)
		return [a[j] for j in sorted(list(a))]
